/* $Id: asmbench.c 175 2004-10-08 15:57:00Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gmp.h>

#include "manu.h"
#include "asm/asm.h"

#define rdtsc(timer) \
     __asm__ __volatile__("nop\nnop\nnop\nnop\nrdtsc" : "=A" (timer))

#define REP2(X)	X X X X
#define REP3(X)	REP2(X) REP2(X)
#define REP6(X) REP3(REP3(X))
#define REP4(X) REP2(REP2(X))
#define REP8(X) REP4(REP4(X))

#define LOOP_SIZE	18
#define INNER_REPEATS	4

static unsigned long long t1,t2;
static long long s,compensate;
static int i;

#define	CALL_MUL1(f,c,a,b)	f(c,a,b)
#define	CALL_MUL2(f,c,a,b)	f(c,a,b)
#define	CALL_MUL3(f,c,a,b)	f(c,a,b)
#define	CALL_MUL4(f,c,a,b)	f(c,a,b)
#define	CALL_MUL5(f,c,a,b)	f(c,a,b)
#define	CALL_MUL6(f,c,a,b)	f(c,a,b)
#define	CALL_MUL7(f,c,a,b)	f(c,a,b)
#define	CALL_MUL8(f,c,a,b)	f(c,a,b)
#define	CALL_ADDMUL1(f,c,a,b)	f(c,a,b)
#define	CALL_ADDMUL2(f,c,a,b)	f(c,a,b)
#define	CALL_ADDMUL3(f,c,a,b)	f(c,a,b)
#define	CALL_ADDMUL4(f,c,a,b)	f(c,a,b)
#define	CALL_SQR1(f,c,a,b)	f(c,a)
#define	CALL_SQR2(f,c,a,b)	f(c,a)
#define	CALL_SQR3(f,c,a,b)	f(c,a)
#define	CALL_SQR4(f,c,a,b)	f(c,a)
#define	CALL_SQR1c(f,c,a,b)	f(c,a)
#define	CALL_SQR2c(f,c,a,b)	f(c,a)
#define	CALL_SQR3c(f,c,a,b)	f(c,a)
#define	CALL_SQR4c(f,c,a,b)	f(c,a)

#define TIMING(t,op,r,k)			\
	mpn_random2(a,8);		\
	mpn_random2(b,8);		\
	s=compensate>>r;		\
	for(i=0;i<(1<<(LOOP_SIZE-r));i++) {	\
		rdtsc(t1);		\
		REP4(CALL_ ## t(op,c,a,b);)	\
		rdtsc(t2);		\
		s+=t2-t1;		\
	}				\
	printf(# op ": %.2f cycles (%.2f/mul)\n",	\
			((double)s)/((double)(1<<(LOOP_SIZE + INNER_REPEATS - r))),	\
			((double)s)/((double)(1<<(LOOP_SIZE + INNER_REPEATS - r))) / k);


#define	DECLARE_TIMER_FUNC(t,op,r,k)	\
    static void _timer_for_ ## op(a,b,c)	\
    	mp_limb_t * a, * b, * c;	\
{					\
	TIMING(t,op,r,k);			\
}

#define	DECLARE_FUNCTION_WRAPPER(t,op)	\
static void CALL_ ## t(fun_ ## op,mp_limb_t * c,mp_limb_t * a,mp_limb_t * b)	\
	__attribute__ ((noinline));		\
static void CALL_ ## t(fun_ ## op,mp_limb_t * c,mp_limb_t * a,mp_limb_t * b)	\
{					\
	CALL_ ## t(op,c,a,b);		\
}					\

#define CALL_TIMER_FUNC(op)	_timer_for_ ## op (a,b,c);

#define	MUL1__R		0
#define	MUL2__R		0
#define	MUL3__R		1
#define	MUL4__R		1
#define	MUL5__R		2
#define	MUL6__R		3
#define	MUL7__R		4
#define	MUL8__R		5
#define	ADDMUL1__R	0
#define	ADDMUL2__R	0
#define	ADDMUL3__R	1
#define	ADDMUL4__R	1
#define	SQR1__R		0
#define	SQR2__R		0
#define	SQR3__R		1
#define	SQR4__R		1
#define	SQR1c__R	0
#define	SQR2c__R	0
#define	SQR3c__R	1
#define	SQR4c__R	1

#define MUL1__NMULS	1
#define MUL2__NMULS	3
#define MUL3__NMULS	6
#define MUL4__NMULS	10
#define MUL5__NMULS	15
#define MUL6__NMULS	21
#define MUL7__NMULS	28
#define MUL8__NMULS	36
#define ADDMUL1__NMULS	1
#define ADDMUL2__NMULS	3
#define ADDMUL3__NMULS	6
#define ADDMUL4__NMULS	10
#define SQR1__NMULS	1
#define SQR2__NMULS	2
#define SQR3__NMULS	4
#define SQR4__NMULS	6
#define SQR1c__NMULS	1
#define SQR2c__NMULS	2
#define SQR3c__NMULS	4
#define SQR4c__NMULS	6

#undef ASM_FUNCTION

#define	ASM_FUNCTION(t,f)		\
	DECLARE_TIMER_FUNC(t,t ## f,t ## __R, t ## __NMULS)	\
	DECLARE_FUNCTION_WRAPPER(t,t ## f)	\
	DECLARE_TIMER_FUNC(t,fun_ ## t ## f,t ## __R, t ## __NMULS)

#include "asm/funcs.h"
#undef ASM_FUNCTION

void bench() {
	mp_limb_t a[8], b[8], c[16];

	/* First do a void computation. */
	s=0LL;
	for(i=0;i<(1<<LOOP_SIZE);i++) {
		rdtsc(t1);
		rdtsc(t2);
		s-=t2-t1;
	}
	printf("Compensation: %.2f cycles\n",((double)-s)/((double)(1<<LOOP_SIZE)));
	compensate=s;

#define	ASM_FUNCTION(t,f)	CALL_TIMER_FUNC(t ## f)
#include "asm/funcs.h"
#undef ASM_FUNCTION

#define	ASM_FUNCTION(t,f)	CALL_TIMER_FUNC(fun_ ## t ## f)
#include "asm/funcs.h"
#undef ASM_FUNCTION
}


int main()
{
	bench();

	return 0;
}
