/* $Id: cxxbench.cpp 182 2004-10-12 17:25:18Z /CN=thome/emailAddress=thome@loria.fr $ */


#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/types.h>

#define	MANU()	(getuid()==6303 || getuid() == 1508 || getuid() == 1001)


#include <gmp.h>
// #include <gmp/gmp-impl.h>

#include <assert.h>

#include "manu.h"

#include "mploc.h"
#include "z2.h"
#include "zp.h"
#include "zpx.h"
#include "z2x.h"
#include "z2n.h"
#include "zpn.h"

#include "timings.h"

#include <iostream>

#include "cxx/z2_cxx.hpp"
#include "cxx/zp_cxx.hpp"
#include "cxx/z2n_cxx.hpp"
#include "cxx/zpn_cxx.hpp"

using namespace std;

gmp_randstate_t prng;

mpz_t random_seed;
#define gmp_getseed(s,prng)     mpz_set(random_seed,prng->_mp_seed)

#define CORRECTNESS_NTESTS      100

#define RANDOM_ELEMENT_PRC      5

int correctness_prc	= 32;
int correctness_degree	= 200;

mp_limb_t random_ui(gmp_randstate_t prng, mp_limb_t m)
{
	mpz_t x;
	mp_limb_t r;
	if (m==0 || m==(mp_limb_t) -1) return 0;
	// otherwise mpz_urandomm crashes
	mpz_init_set_ui(x,m);
	mpz_urandomm(x,prng,x);
	r=mpz_get_ui(x);
	mpz_clear(x);
	return r;
}

// #define random_degree	random_ui(prng,correctness_degree)
#define random_degree	correctness_degree

/* Easier to put a breakpoint on */
static void failure_exit(void)
{
	exit(255);					\
}

#define CHECKTEST(msg,p) /* moan if a!=b, or precision less than p. */	\
	BEGIN_BLOCK							\
		if (a.x->prec<p || b.x->prec <p || !(a==b))	{	\
			ok=0;						\
			cerr << "TEST " << i << " FAILED ";		\
			cerr << "["  << msg << "]\n";			\
			cerr << "x:=" << x << ";\n";			\
			cerr << "y:=" << y << ";\n";			\
			cerr << "z:=" << z << ";\n";			\
			cerr << "a:=" << a << ";\n";			\
			cerr << "b:=" << b << ";\n";			\
			cerr << "c:=" << c << ";\n";			\
			failure_exit();					\
		}							\
	END_BLOCK
		
/* These tests for zero-precision behavior are meant only to trigger
 * corner-case segfaults.
 */
template<class T> void zero_ring_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing zero-precision\n";

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		T::random(x,prng,random_degree,0);
		T::random(y,prng,random_degree,0);
		T::random(z,prng,random_degree,0);

		T::add(a,x,y);
		T::add(b,x,x);
		T::mul(c,x,y);
		T::mul(z,z,y);

		T::set(z,a);
		T::set(b,a);

		a.set(random_ui(prng,CORRECTNESS_NTESTS/2),0);
		a.set(random_ui(prng,CORRECTNESS_NTESTS/2),1);

		z.valuation();
	}

	pop_time("Time for zero-width tests");
}

template<class T> void add_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing addition\n";
	int ok=1;

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);
		
		T::random(x,prng,random_degree,RANDOM_ELEMENT_PRC);
		T::random(y,prng,random_degree,RANDOM_ELEMENT_PRC);
		T::random(z,prng,random_degree,RANDOM_ELEMENT_PRC);

		x.force_precision(correctness_prc);
		y.force_precision(correctness_prc);
		z.force_precision(correctness_prc);
		
		T::add(a,x,y);
		T::add(b,a,z);
		T::add(c,y,z);
		T::add(a,x,c);
		CHECKTEST("addition associativity",correctness_prc);

		T::add(a,x,y);
		T::add(a,a,z);
		T::add(b,y,z);
		T::add(b,b,x);
		CHECKTEST("add. assoc. w/ left overlap",correctness_prc);

		T::add(a,x,y);
		T::add(a,z,a);
		T::add(b,y,z);
		T::add(b,x,b);
		CHECKTEST("add. assoc. w/ right overlap",correctness_prc);

		T::add(a,y,x);
		T::add(b,x,y);
		CHECKTEST("add. commutativity",correctness_prc);

		T::sub(a,y,x);
		T::neg(c,x);
		T::add(b,y,c);
		CHECKTEST("subtract + neg",correctness_prc);

		T::add(a,y,x);
		T::neg(c,x);
		T::sub(b,y,c);
		CHECKTEST("subtract + 2neg",correctness_prc);
	}

	pop_time("Time for addition tests");
}

template<class T> void mul_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing multiplication\n";
	int ok=1;

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);
		
		T::random(x,prng,random_degree,RANDOM_ELEMENT_PRC);
		T::random(y,prng,random_degree,RANDOM_ELEMENT_PRC);
		T::random(z,prng,random_degree,RANDOM_ELEMENT_PRC);

		x.force_precision(correctness_prc);
		y.force_precision(correctness_prc);
		z.force_precision(correctness_prc);
		
		T::mul(a,x,y);
		T::mul(b,a,z);
		T::mul(c,y,z);
		T::mul(a,x,c);
		CHECKTEST("mul. associativity",correctness_prc);

		T::mul(a,x,y);
		T::mul(a,a,z);
		T::mul(b,y,z);
		T::mul(b,b,x);
		CHECKTEST("mul. assoc. w/ left overlap",correctness_prc);

		T::mul(a,x,y);
		T::mul(a,z,a);
		T::mul(b,y,z);
		T::mul(b,x,b);
		CHECKTEST("mul. assoc. w/ right overlap",correctness_prc);

		T::mul(a,y,x);
		T::mul(b,x,y);
		CHECKTEST("mul. commutativity",correctness_prc);

		T::mul(b,z,x);
		T::mul(c,z,y);
		T::add(a,b,c);
		T::add(c,x,y);
		T::mul(b,z,c);
		CHECKTEST("mul. left distributivity",correctness_prc);
	}

	pop_time("Time for multiplication tests");
}

template<class T> void sqr_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing squaring\n";
	int ok=1;

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);
		
		T::random(x,prng,random_degree,RANDOM_ELEMENT_PRC);
		T::random(y,prng,random_degree,RANDOM_ELEMENT_PRC);

		x.force_precision(correctness_prc);
		y.force_precision(correctness_prc);
		
		/* Verify that (x-y)^2 + (x+y)^2 = 2*(x^2+y^2) */
		T::sub(z,x,y);
		T::sqr(b,z);
		T::add(z,x,y);
		T::sqr(c,z);
		T::add(a,b,c);
		T::sqr(b,x);
		T::sqr(c,y);
		T::add(b,b,c);
		T::add(b,b,b);	/* Can't use lshift in characteristic != 2 !! */
		CHECKTEST("squaring",correctness_prc);
	}

	pop_time("Time for squaring tests");
}

template<class T> void inv_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing inversion\n";
	int ok=1, fl=0;

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);
		
		T::random(x,prng,RANDOM_ELEMENT_PRC);

		x.force_precision(correctness_prc);

		if (x.valuation()) { fl++; continue; }

		T::inv(y,x);
		T::add(y,y,1);

		if (y.valuation()) { fl++; continue; }

		T::inv(z,y);
		T::add(c,x,1);
		T::mul(b,c,z);
		T::set(a,x);
		CHECKTEST("inversion",correctness_prc);

		/* Contents of registers after test:
			x = x
			y = 1+1/x
			z = 1/y = x/(x+1)
			a = x
			b = (x+1)*x/(x+1)
			c = x+1
		*/
	}

	cout << fl << " bad inputs encoutered\n";
	cout << "In theory : " << CORRECTNESS_NTESTS << "/(car K/2)\n";

	pop_time("Time for inversion tests");
}

template<class T> void invsqrt_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing invsqrt\n";
	int ok=1, fl=0;

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);
		
		T::random(x,prng,RANDOM_ELEMENT_PRC);

		x.force_precision(correctness_prc);

		if (x.valuation()) { fl++; continue; }

		T::sqr(z,x);	// build sth that is sqrt-able...

		if (!T::invsqrt(y,z)) {
			cerr << "Argh. Should have invsqrt...\n";
			BUG();
		}


		T::sqr(a,y);
		T::mul(a,a,z);
		b.set(1,correctness_prc);
		CHECKTEST("invsqrt",correctness_prc);

		/* Contents of registers after test:
			x = x
			y = 1/sqrt(x)
			z = 1/x
			a = x/x
			b = 1
		*/
	}

	cout << fl << " bad inputs encoutered\n";

	if (MANU() &&
		fl > 1.1 * CORRECTNESS_NTESTS / mpz_get_d(parent.cval()->p)) {
		BASH_USER("Fix these many bad inputs\n");
	}


	pop_time("Time for invsqrt tests");
}

template<class T> void sqrt_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing sqrt\n";
	int ok=1;

	push_time();

	for(int i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);
		
		T::random(x,prng,RANDOM_ELEMENT_PRC);

		x.force_precision(correctness_prc);

		int v = x.valuation();

		T::rshift(x,x,v);
		T::sqr(z,x);	// build sth that is sqrt-able...

		T::lshift(x,x,v);
		T::lshift(z,z,2*v);

		if (!T::sqrt(y,z)) {
			cerr << "Argh. Should have sqrt...\n";
			BUG();
		}


		T::set(a,x);
		T::set(b,y);

		/* At this point we should have either a==b or a==-b,
		 * since x can be any of the two square roots of its
		 * square. To avoid such sign jokes, we square again the
		 * results in order to rely on the a==b check.
		 */
		T::sqr(a,a);
		T::sqr(b,b);

		CHECKTEST("sqrt",correctness_prc);
	}

	pop_time("Time for sqrt tests");
}

template<class T> void div_test(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	cout << "Testing Euclidean division\n";

	int ok=1,fl=0;
	push_time();
	for(int i=0;i<CORRECTNESS_NTESTS;i++) {

		gmp_getseed(random_seed,prng);
		gmp_randseed(prng,random_seed);

		T::random(x,prng,random_degree,correctness_prc);
		T::random(y,prng,random_degree,correctness_prc);
		T::random(z,prng,random_degree,correctness_prc);

		x.force_precision(correctness_prc);
		y.force_precision(correctness_prc);
		z.force_precision(correctness_prc);

		if (!y.can_divide()) { fl++; continue; }

		T::set(a,x);
		T::div(z,c,a,y);
		T::mul(b,z,y);
		T::add(b,b,c);
		CHECKTEST("division",correctness_prc);

		T::set(a,x);
		T::div(x,c,x,y);
		T::set(b,x);
		T::mul(b,b,y);
		T::add(b,b,c);
		T::set(x,a);
		CHECKTEST("division w/ quotient overlap",correctness_prc);

		T::set(a,x);
		T::div(b,x,x,y);
		T::mul(b,b,y);
		T::add(b,b,x);
		T::set(x,a);
		CHECKTEST("division w/ remainder overlap",correctness_prc);

		// NO overlap is possible on the divisor.
		
		int k = random_ui(prng,x.deg());
		T::set(a,x);
		T::div_q_xi(b,x,k);
		T::div_r_xi(c,x,k);
		T::mul_xi(b,b,k);
		T::add(b,b,c);
		CHECKTEST("division by powers",correctness_prc);
	}

	cout << fl << " bad inputs encoutered\n";

	pop_time("Time for division tests");
}

template <class T> void field_tests(typename T::field& parent)
{
	zero_ring_test<T>(parent);
	add_test<T>(parent);
	mul_test<T>(parent);
	sqr_test<T>(parent);
	inv_test<T>(parent);
	invsqrt_test<T>(parent);
	sqrt_test<T>(parent);
}

template <class T> void polring_tests(typename T::field& parent)
{
	zero_ring_test<T>(parent);
	add_test<T>(parent);
	mul_test<T>(parent);
	sqr_test<T>(parent);
	div_test<T>(parent);
}

void zp_tests(mpz_ptr p)
{
	/*********************************/
	printf("#### Zp test\n\n\n");

	Mploc_zp_field fzp(p);
	field_tests<zp_elt>(fzp);
}

void zpx_tests(mpz_ptr p)
{
	//*********************************/
	printf("#### ZpX test\n\n\n");

	Mploc_zp_field fzp(p);
	polring_tests<zp_poly>(fzp);
}


void z2_tests()
{
	/*********************************/
	printf("#### Z2 test\n\n\n");

	Mploc_z2_field fz2;

	field_tests<z2_elt>(fz2);
}

void z2n_tests()
{
	/*********************************/
	printf("#### Z2n test\n\n\n");

	Mploc_z2_field fz2;
	z2_poly p(fz2);
	p.set_xi(6,1000);
	z2_poly::add(p,p,1);
	z2_poly::mul_xi(p,p,1);
	z2_poly::add(p,p,1);	/* X^7+X+1 */

	Mploc_z2n_field fz2n(p);

	field_tests<z2n_elt>(fz2n);
}

void zpn_tests(mpz_ptr q)
{
	/*********************************/
	printf("#### zpn test\n\n\n");

	Mploc_zp_field fzp(q);
	zp_poly p(fzp);
	p.set_xi(6,1000);
	zp_poly::add(p,p,1);
	zp_poly::mul_xi(p,p,1);
	zp_poly::add(p,p,1);	/* X^7+X+1 */

	Mploc_zpn_field fzpn(p);

	field_tests<zpn_elt>(fzpn);
}

void z2x_tests()
{
	//*********************************/
	printf("#### Z2X test\n\n\n");

	Mploc_z2_field fz2;

	polring_tests<z2_poly>(fz2);
}


int main(int argc, char * argv[])
{
	mpz_t	p UNUSED_VARIABLE;
	int	i UNUSED_VARIABLE;

	setvbuf(stdout,NULL,_IONBF,0);
	setvbuf(stderr,NULL,_IONBF,0);

	mpz_init(random_seed);

	gmp_randinit_default(prng);

	for(;argc>=3;) {
		if (argc>=3 && strcmp(argv[1],"-d")==0) {
			correctness_degree=atoi(argv[2]);
			argc-=2;
			argv+=2;
			continue;
		}
		if (argc>=3 && strcmp(argv[1],"-p")==0) {
			correctness_prc=atoi(argv[2]);
			argc-=2;
			argv+=2;
			continue;
		}
		break;
	}
	
	/* If the user gives arguments for running zp tests, he probably
	 * wants to have them run first.
	 */
	if (argc >= 3) {
		mpz_init_set_ui(p,atoi(argv[1]));
		mpz_pow_ui(p,p,atoi(argv[2]));
		for(i=0;!mpz_probab_prime_p(p,5);i++) mpz_add_ui(p,p,1);
		printf("p=%d^%d + %d\n", atoi(argv[1]), atoi(argv[2]), i);

		zp_tests(p);
		zpx_tests(p);
		zpn_tests(p);
	}

	z2_tests();
	z2x_tests();
	z2n_tests();

	return 0;
}

