/* $Id: cxxtimer.cpp 169 2004-09-30 14:39:30Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <sys/types.h>

#include <gmp.h>

#include <assert.h>

#include "manu.h"

#include "mploc.h"
#include "z2.h"
#include "zp.h"
#include "zpx.h"
#include "z2x.h"
#include "z2n.h"
#include "zpn.h"

#include "timings.h"
#include "irr_tbl.h"

#include <iostream>

#define	CTORLOG(x)	/**/

#include "cxx/z2_cxx.hpp"
#include "cxx/zp_cxx.hpp"
#include "cxx/z2n_cxx.hpp"
#include "cxx/zpn_cxx.hpp"

using namespace std;

gmp_randstate_t prng;

int ntests = 1000;
int max_check_time_ms = 500;
int degree_for_poly_tests = 32;

#define	BORED	(max_check_time_ms &&	\
		(look_time_as_sec() > (1e-3 * (double)max_check_time_ms)))

static int precisions_to_test[] UNUSED_VARIABLE ={
	1,2,
	3,4,5,6,
	7,8,9,12,
	15,16,17,
	23,24,25,
	31,32,33,
	47,48,49,
	63,64,65,
	95,96,97,
	127,128,129,
	159,160,161,
	191,192,193,
	223,224,225,
	255,256,257,
	287,288,289,
	319,320,321,
	351,352,353,
	383,384,385,
	415,416,417,
	447,448,449,
	479,480,481,
	511,512,513,
	1024,2048,4096,8192,16384,32768
};

static int seeds_to_test[] UNUSED_VARIABLE ={
	1,2,
	3,4,6,
	7,8,12,
	15,16,24,
	31,32,48,
	63,64,96,
	127,128,192,
	255,256,384,
	511,512};

#define	NPRC	(sizeof(precisions_to_test)/sizeof(mp_size_t))
#define	NSEEDS	(sizeof(seeds_to_test)/sizeof(int))

template<class T> void	bench_mul(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);
	for(unsigned int j=0;j<NPRC;j++) {
		mp_size_t curr_prec = precisions_to_test[j];

		/* the degree parameter is ignored for polynomial rings */
		T::random(x,prng,degree_for_poly_tests,curr_prec);
		T::random(y,prng,degree_for_poly_tests,curr_prec);

		push_time();
		for(int i=0;i<ntests && !BORED;i++) {
			T::mul(z,x,y);
			x.normalize();
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}
}

template<class T> void	bench_sqr(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);
	for(unsigned int j=0;j<NPRC;j++) {
		mp_size_t curr_prec = precisions_to_test[j];

		/* the degree parameter is ignored for polynomial rings */
		T::random(x,prng,degree_for_poly_tests,curr_prec);

		push_time();
		for(int i=0;i<ntests && !BORED;i++) {
			T::sqr(y,x);
			y.normalize();
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}
}

template<class T> void	bench_inv(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	for(unsigned int j=1;j<NPRC;j++) {
		mp_size_t curr_prec = precisions_to_test[j];

		do { T::random(x,prng,curr_prec); } while (x.valuation());

		push_time();
		for(int i=0;i<ntests && !BORED;i++) {
			T::inv(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}
}

template<class T> void	bench_seeded_inv(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	for(unsigned int j=0;j<NSEEDS;j++) {
		T::field::inversion_configure(1,seeds_to_test[j]);
		push_time();

		for(int i=0;i<ntests && !BORED;i++) {
			T::random(x,prng,precisions_to_test[NPRC-1]);
			if (x.valuation()) continue;

			T::inv(y,x);
		}
		printf("%d ",seeds_to_test[j]);
		pop_time_raw();
		printf("\n");
	}
}

template<class T> void	bench_invsqrt(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	for(unsigned int j=1;j<NPRC;j++) {
		mp_size_t curr_prec = precisions_to_test[j];

		do { T::random(x,prng,curr_prec); } while (x.valuation());
		T::sqr(x,x);	/* Make sure we've got a sqr ! */

		push_time();
		for(int i=0;i<ntests && !BORED;i++) {
			T::invsqrt(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}
}

template<class T> void	bench_sqrt(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);

	for(unsigned int j=1;j<NPRC;j++) {
		mp_size_t curr_prec = precisions_to_test[j];

		T::random(x,prng,curr_prec);
		T::sqr(x,x);	/* Make sure we've got a sqr ! */

		push_time();
		for(int i=0;i<ntests && !BORED;i++) {
			T::sqrt(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}
}

template<class T> void	var_width_field_timings(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);
	
	timings_set_dispmode(TIMINGS_ARE_SECONDS);

	cout << "Variable-width timings\n";

	cout << "multiplication\n";
	bench_mul<T>(parent);

	cout << "squaring\n";
	bench_sqr<T>(parent);

	cout << "full-extent inversion (iterative)\n";
	T::field::inversion_configure(0,0);
	bench_inv<T>(parent);

	cout << "full-extent inversion (recursive)\n";
	T::field::inversion_configure(1,0);
	bench_inv<T>(parent);

	cout << "seeded inversion at precision "
		<< precisions_to_test[NPRC-1]
		<< " (recursive)\n";
	bench_seeded_inv<T>(parent);

	/* We return to standard recursive code */
	T::field::inversion_configure(1,0);

	cout << "invsqrt\n";
	bench_invsqrt<T>(parent);

	cout << "sqrt\n";
	bench_sqrt<T>(parent);

}

template<class T> void	var_width_ring_timings(typename T::field & parent)
{
	T x(parent), y(parent), z(parent);
	T a(parent), b(parent), c(parent);
	
	timings_set_dispmode(TIMINGS_ARE_SECONDS);

	cout << "Variable-width timings\n";

	/**********************************************************/
	cout << "multiplication\n";
	bench_mul<T>(parent);

	/**********************************************************/
	cout << "squaring\n";
	bench_sqr<T>(parent);
}

template <class T> void field_tests(typename T::field& parent)
{
	var_width_field_timings<T>(parent);
}

template <class T> void polring_tests(typename T::field& parent)
{
	var_width_ring_timings<T>(parent);
}

void zp_tests(mpz_ptr p)
{
	/*********************************/
	printf("#### Zp test\n\n\n");

	Mploc_zp_field fzp(p);
	field_tests<zp_elt>(fzp);
}

void zpx_tests(mpz_ptr p)
{
	//*********************************/
	printf("#### ZpX test\n\n\n");

	Mploc_zp_field fzp(p);
	polring_tests<zp_poly>(fzp);
}


void z2_tests()
{
	/*********************************/
	printf("#### Z2 test\n\n\n");

	Mploc_z2_field fz2;

	field_tests<z2_elt>(fz2);
}

void z2n_tests()
{
	/*********************************/
	printf("#### Z2n test\n\n\n");

	Mploc_z2_field fz2;
	z2_poly p(fz2);
	p.set_xi(6,1000);
	z2_poly::add(p,p,1);
	z2_poly::mul_xi(p,p,1);
	z2_poly::add(p,p,1);	/* X^7+X+1 */

	Mploc_z2n_field fz2n(p);

	field_tests<z2n_elt>(fz2n);
}

void zpn_tests(mpz_ptr q)
{
	/*********************************/
	printf("#### zpn test\n\n\n");

	Mploc_zp_field fzp(q);
	zp_poly p(fzp);
	p.set_xi(6,1000);
	zp_poly::add(p,p,1);
	zp_poly::mul_xi(p,p,1);
	zp_poly::add(p,p,1);	/* X^7+X+1 */

	Mploc_zpn_field fzpn(p);

	field_tests<zpn_elt>(fzpn);
}

void z2x_tests()
{
	//*********************************/
	printf("#### Z2X test\n\n\n");

	Mploc_z2_field fz2;

	polring_tests<z2_poly>(fz2);
}

void usage(int rc)
{
	cerr << "Usage: cxxtimer [options] <module> [<a> <b>] <func>\n";
	cerr << "Give a,b only for the zp case (p is then nextprime(a^b))\n";
	cerr << "Accepted options:\n";
	cerr << "\t--help:\tDisplay this help\n";
	cerr << "\t--deg <n>:\tFor z2n, zpn, define degree of extension\n";
	cerr << "\t--tm <time>:\tMaximum number of milliseconds per test\n";
	cerr << "\t--ts <time>:\tMaximum number of elementary tests\n";
	exit(rc);
}

int main(int argc, char * argv[])
{
	int	t;

	setvbuf(stdout,NULL,_IONBF,0);
	setvbuf(stderr,NULL,_IONBF,0);

	gmp_randinit_default(prng);

	for(t=0;argc>1;argv++,argc--) {
		if (strcmp(argv[1],"--help")==0) usage(0);
		if (strcmp(argv[1],"--tm")==0) {
			if (argc<=2) usage(1);
			max_check_time_ms = atoi(argv[2]);
			argc++,argv--;
			continue;
		}
		if (strcmp(argv[1],"--ts")==0) {
			if (argc<=2) usage(1);
			ntests = atoi(argv[2]);
			argc++,argv--;
			continue;
		}
	}

	z2_tests();
	z2x_tests();
	z2n_tests();

	/* Creation ; Selection of prime number. */
	if (argc >= 3) {
		mpz_t	p;
		int i;
		mpz_init_set_ui(p,atoi(argv[1]));
		mpz_pow_ui(p,p,atoi(argv[2]));
		for(i=0;!mpz_probab_prime_p(p,5);i++) mpz_add_ui(p,p,1);
		printf("p=%d^%d + %d\n", atoi(argv[1]), atoi(argv[2]), i);

		zp_tests(p);
		zpx_tests(p);
		zpn_tests(p);
	}

	return 0;
}
