/* $Id: asm001.c 205 2006-09-29 14:42:37Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gmp.h>

#include "manu.h"

#include "z2_limbs.h"

#include "asm/asm.h"

#undef  ASM_FUNCTION

int some_failed=0;
#define ASM_FUNCTION(t,f)	\
	int failed_ ## t ## f = 0;
#include "asm/funcs.h"
#undef	ASM_FUNCTION

#define	VERIFY_WORDS(f,c,t,w,i,v)		\
	BEGIN_BLOCK				\
	for(i=0;i<w;i++) {			\
		if (c[i]==t[i]) continue;	\
		if (failed_ ## f++ < 10)	\
		fprintf(stderr,			\
			"(" #f ") [%d] : %08lx != %08lx (test %d)\n",	\
			i,c[i],t[i],v); 	\
		some_failed=1;			\
		break;				\
	}					\
	END_BLOCK

#define MUL_run(f,w)				\
	BEGIN_BLOCK				\
	mpn_random2(a,w);			\
	mpn_random2(b,w);			\
	mpn_mul_n(t,a,b,w);			\
	f(c,a,b);				\
	{					\
		int i;				\
		VERIFY_WORDS(f,c,t,w,i,v);	\
	}					\
	END_BLOCK

#define SQR_run(f,w)				\
	BEGIN_BLOCK				\
	mpn_random2(a,w);			\
	mpn_mul_n(t,a,a,w);			\
	f(c,a);					\
	{					\
		int i;				\
		VERIFY_WORDS(f,c,t,w,i,v);	\
	}					\
	END_BLOCK

#define SQRc_run(f,w)				\
	BEGIN_BLOCK				\
	mp_limb_t r;				\
	mpn_random2(a,w);			\
	mpn_mul_n(t,a,a,w);			\
	r=f(c,a);				\
	{					\
		int i;				\
		VERIFY_WORDS(f,c,t,w,i,v);	\
		if ((r&1UL)!=(t[w]&1UL)) {	\
			if (failed_ ## f++ < 10)	\
			fprintf(stderr,		\
			"(" #f ") [carry] : %08lx != %08lx (test %d)\n",\
					r&1UL,t[w]&1UL,v);	\
			some_failed=1;		\
			break;			\
		}				\
	}					\
	END_BLOCK

#define MUL_run_overlap1(f,w)			\
	BEGIN_BLOCK				\
	mpn_random2(a,w);			\
	mpn_random2(b,w);			\
	mpn_mul_n(t,a,b,w);			\
	f(a,a,b);				\
	{					\
		int i;				\
		VERIFY_WORDS(f ## _OV1,a,t,w,i,v);	\
	}					\
	END_BLOCK

#define MUL_run_overlap2(f,w)			\
	BEGIN_BLOCK				\
	mpn_random2(a,w);			\
	mpn_random2(b,w);			\
	mpn_mul_n(t,a,b,w);			\
	f(a,a,b);				\
	{					\
		int i;				\
		VERIFY_WORDS(f ## _OV2,a,t,w,i,v);	\
	}					\
	END_BLOCK

#define SQR_run_overlap(f,w)			\
	BEGIN_BLOCK				\
	mpn_random2(a,w);			\
	mpn_mul_n(t,a,a,w);			\
	f(a,a);					\
	{					\
		int i;				\
		VERIFY_WORDS(f ## _OV,a,t,w,i,v);	\
	}					\
	END_BLOCK

#define SQRc_run_overlap(f,w)				\
	BEGIN_BLOCK				\
	mp_limb_t r;				\
	mpn_random2(a,w);			\
	mpn_mul_n(t,a,a,w);			\
	r=f(a,a);				\
	{					\
		int i;				\
		VERIFY_WORDS(f ## _OV,a,t,w,i,v);	\
		if ((r&1UL)!=(t[w]&1UL)) {	\
			if (failed_ ## f++ < 10)	\
			fprintf(stderr,		\
			"(" #f ") [carry] : %08lx != %08lx (test %d)\n",\
					r&1UL,t[w]&1UL,v);	\
			some_failed=1;		\
			break;			\
		}				\
	}					\
	END_BLOCK

#define ADDMUL_run(f,w)				\
	BEGIN_BLOCK				\
	mpn_random2(a,w);			\
	mpn_random2(b,w);			\
	mpn_random2(c,w);			\
	memcpy(d,c,w*sizeof(mp_limb_t));	\
	mpn_mul_n(t,a,b,w);			\
	mpn_add_n(t,t,d,w);			\
	f(c,a,b);				\
	{					\
		int i;				\
		VERIFY_WORDS(f,c,t,w,i,v);	\
	}					\
	END_BLOCK

#define	CHECK_MUL1(t,f)		MUL_run(f, 1)
#define	CHECK_MUL2(t,f)		MUL_run(f, 2)
#define	CHECK_MUL3(t,f)		MUL_run(f, 3)
#define	CHECK_MUL4(t,f)		MUL_run(f, 4)
#define	CHECK_MUL5(t,f)		MUL_run(f, 5)
#define	CHECK_MUL6(t,f)		MUL_run(f, 6)
#define	CHECK_MUL7(t,f)		MUL_run(f, 7)
#define	CHECK_MUL8(t,f)		MUL_run(f, 8)
#define	CHECK_ADDMUL1(t,f)	ADDMUL_run(f, 1)
#define	CHECK_ADDMUL2(t,f)	ADDMUL_run(f, 2)
#define	CHECK_ADDMUL3(t,f)	ADDMUL_run(f, 3)
#define	CHECK_ADDMUL4(t,f)	ADDMUL_run(f, 4)
#define	CHECK_SQR1(t,f)		SQR_run(f, 1)
#define	CHECK_SQR2(t,f)		SQR_run(f, 2)
#define	CHECK_SQR3(t,f)		SQR_run(f, 3)
#define	CHECK_SQR4(t,f)		SQR_run(f, 4)
#define	CHECK_SQR1c(t,f)	SQRc_run(f, 1)
#define	CHECK_SQR2c(t,f)	SQRc_run(f, 2)
#define	CHECK_SQR3c(t,f)	SQRc_run(f, 3)
#define	CHECK_SQR4c(t,f)	SQRc_run(f, 4)


void test(int v)
{
	mp_limb_t a[8], b[8], c[8], d[8], t[16];
#define	ASM_FUNCTION(t,f)	\
	if (!failed_ ## t ## f) {	\
		CHECK_ ## t(t,t ##f);	\
	}
#include "asm/funcs.h"
#undef	ASM_FUNCTION
}

#define	MAXW	32

#define	FUNC_SP_WRAP(c,a,b)	mploc_limbs_short_product(c,a,b,tmp,w)
#define	FUNC_SQ_WRAP(c,a)	mploc_limbs_short_square(c,a,tmp,w)
#define	FUNC_SQC_WRAP(c,a)	mploc_limbs_short_square_carry(c,a,tmp,w)

int failed_FUNC_SP_WRAP=0;
int failed_FUNC_SQ_WRAP=0;
int failed_FUNC_SQC_WRAP=0;
int failed_FUNC_SP_WRAP_OV1=0;
int failed_FUNC_SP_WRAP_OV2=0;
int failed_FUNC_SQ_WRAP_OV=0;
int failed_FUNC_SQC_WRAP_OV=0;

void wtest(int v)
{
	mp_limb_t a[MAXW], b[MAXW], c[MAXW], t[2*MAXW], tmp[2*MAXW];
	mp_size_t w;

	for(w=1;w<MAXW;w++) {
		mpn_random2(a,w);
		mpn_random2(b,w);

		MUL_run(FUNC_SP_WRAP,w);
		SQR_run(FUNC_SQ_WRAP,w);
		SQRc_run(FUNC_SQC_WRAP,w);

		MUL_run_overlap1(FUNC_SP_WRAP,w);
		MUL_run_overlap2(FUNC_SP_WRAP,w);
		SQR_run_overlap(FUNC_SQ_WRAP,w);
		SQRc_run_overlap(FUNC_SQC_WRAP,w);
	}
}
	


int main(int argc, char * argv[])
{
	int k;
	int i;
	int verbose = 0;
	for(i = 1 ; i < argc ; i++) {
		if (strcmp(argv[i], "-v") == 0) { verbose=1; continue; }
		fprintf(stderr, "Bad option: %s\n", argv[i]);
		exit(1);
	}

	for(k=0;k<10000;k++) {
		test(k);
	}

	for(k=0;k<100;k++) {
		wtest(k);
	}

	if (verbose || some_failed) {
#define	ASM_FUNCTION(t,f)	\
	printf(# t # f ": %s\n", failed_ ## t ## f ? "BAD" : "ok");
#include "asm/funcs.h"
	ASM_FUNCTION(FUNC,_SP_WRAP)
	ASM_FUNCTION(FUNC,_SQ_WRAP)
	ASM_FUNCTION(FUNC,_SQC_WRAP)
	ASM_FUNCTION(FUNC,_SP_WRAP_OV1)
	ASM_FUNCTION(FUNC,_SP_WRAP_OV2)
	ASM_FUNCTION(FUNC,_SQ_WRAP_OV)
	ASM_FUNCTION(FUNC,_SQC_WRAP_OV)
	}

	return some_failed;
}
