/* $Id: agm2.c 196 2004-10-14 16:07:44Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <string.h>
#include <stdlib.h>

#include "mploc.h"
#include "z2.h"
#include "z2x.h"
#include "z2n.h"

#include "timings.h"
#include "irr_tbl.h"

#define	DEFAULT_DEG		163

mplf_t z2,z2n;
mploc_z2n_t t0,t1;

const int a6_coeffs[]={
	0, 3, 5, 8, 9, 10, 12, 13, 15, 16, 17, 18, 19, 20, 21, 22, 25,
	29, 33, 36, 38, 39, 40, 41, 44, 47, 48, 49, 51, 55, 57, 60, 61,
	62, 63, 64, 67, 70, 71, 72, 73, 74, 77, 78, 79, 84, 87, 88, 89,
	90, 91, 94, 100, 102, 103, 104, 106, 107, 108, 111, 115, 117,
	124, 128, 129, 130, 131, 132, 134, 136, 141, 142, 145, 146, 150,
	154, 155, 158, 161};

void usage(int rc)
{
	fprintf(stderr, "Usage: agm <deg>\n");
	exit(rc);
}

void biAGM_step(mploc_z2n_ptr a, mploc_z2n_ptr b, mp_size_t k)
{
	mploc_z2n_force_precision(a, k + 1);
	mploc_z2n_force_precision(b, k + 1);
	mploc_z2n_add(t0, a, b);
	mploc_z2n_rshift(t0, t0, 1);
	mploc_z2n_mul(t1, a, b);
	mploc_z2n_sqrt(t1, t1);

	mploc_z2n_set(a, t0);

	/* Make sure the sqrt is chosen of the form 1+4u */
	mploc_z2n_sub_ui(t0, t1, 1);
	if (mploc_z2n_valuation(t0) < 2)
		mploc_z2n_neg(t1, t1);

	mploc_z2n_set(b, t1);
}

void singleAGM_step(mploc_z2n_ptr lambda, mp_size_t k)
{
	mploc_z2n_force_precision(lambda, k + 1);
	mploc_z2n_add_ui(t0, lambda, 1);
	mploc_z2n_rshift(t0, t0, 1);
	mploc_z2n_invsqrt(lambda, lambda);
	mploc_z2n_mul(lambda, lambda, t0);
}

void BivariateAGM(mploc_z2n_t res, mploc_z2n_t a6)
{
	mploc_z2n_t a, b, a0;
	int k, maxk;
	int i;
	int prec;

	mploc_z2n_init(a, z2n);
	mploc_z2n_init(a0, z2n);
	mploc_z2n_init(b, z2n);

	maxk = (z2n->degree) / 2 + 4;
//  maxk = 2*maxk;
	prec = maxk + 1;

	/* init: a = 1 + 8*a6, b = 1 mod 16 */
	mploc_z2n_lshift(a, a6, 3);
	mploc_z2n_add_ui(a, a, 1);
	mploc_z2n_set_ui(b, 1, prec);

	/* AGM loop */
	printf("precision\n");
	for (k = 4; k <= maxk; k++) {
		biAGM_step(a,b,k+1);
	}

	mploc_z2n_set(a0, a);

	/* one more complete loop */
	printf("loop\n");
	for (i = 0; i < z2n->degree; i++) {
		biAGM_step(a,b,maxk+1);
	}

	/* return the quotient a0/a */
	mploc_z2n_inv(a, a);
	mploc_z2n_mul(res, a, a0);


	mploc_z2n_clear(a);
	mploc_z2n_clear(a0);
	mploc_z2n_clear(b);
}


void AGM(mploc_z2n_ptr a6)
{
	mploc_z2n_t lambda, aux;
	int k, maxk;

	mploc_z2n_init(lambda, z2n);
	mploc_z2n_init(aux, z2n);

	maxk = (z2n->degree) / 2 + 4;

	/* lambda := 1 + 8*a6 */
	mploc_z2n_lshift(lambda, a6, 3);
	mploc_z2n_add_ui(lambda, lambda, 1);

	/* AGM loop:  lambda' := (1+lambda) / (2*sqrt(lambda)) */
	printf("precision\n");
	for (k = 4; k <= maxk; k++) {
		singleAGM_step(lambda,k+1);
	}

	/* Compute Norm(2*lambda/(1+lambda)) */

	mploc_z2n_clear(lambda);
	mploc_z2n_clear(aux);

}

/* convert an element of Z_q into a signed integer in the Hasse-Weil
 * interval */
void printTraceInHW(mploc_z2n_t tr)
{
	mploc_t Tr;
	mplf_t z2;
	mpz_t zTr, modulus;
	int HWlim, sign;

	HWlim = (z2n->degree) / 2 + 4;

	z2 = mplf_z2_create();
	mploc_z2_init(Tr, z2);

	mploc_z2x_get_coeff(Tr, tr, 0);
	mpz_init(zTr);
	mpz_set(zTr, Tr->x);
	mpz_init(modulus);
	mpz_set_ui(modulus, 1);
	mpz_mul_2exp(modulus, modulus, HWlim);
	mpz_mod(zTr, zTr, modulus);
	mpz_set_ui(modulus, 1);
	mpz_mul_2exp(modulus, modulus, HWlim - 1);
	if (mpz_cmp(zTr, modulus) > 0) {
		sign = -1;
		mpz_mul_ui(modulus, modulus, 2);
		mpz_sub(zTr, modulus, zTr);
	} else {
		sign = 1;
	}

	if (sign == -1)
		printf("-");
	mpz_out_str(stdout, 10, zTr);

	mpz_clear(modulus);
	mpz_clear(zTr);
	mploc_z2_clear(Tr);
	mplf_z2_destroy(&z2);
}


int main(int argc, char * argv[])
{
	mploc_z2x_t f;
	mplf_t z2;
	gmp_randstate_t prng;
	int p;
	int deg=DEFAULT_DEG;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	gmp_randinit_default(prng);

	/* create Z_2 */
	z2 = mplf_z2_create();
	/* create and initialize f */
	mploc_z2x_init(f, z2);	/* necessary to build z2 first??? */

	for(p=0;argc>1;argv++,argc--) {
		if (strcmp(argv[1],"--help")==0) {
			usage(0);
		}
		mp_size_t v=atoi(argv[1]);
		if (v==0) usage(1);
		switch (p++) {
			case 0: deg=v; break;
			default: usage(1);
		}
	}

	if (!mploc_z2x_set_irrpol_db_quick(f,deg)) {
		fprintf(stderr,"Degree %d poly is not in irr_tbl\n", deg);
		exit(1);
	}
		
	mploc_z2x_force_precision(f,deg/2+5);

	printf("f=");
	mploc_z2x_print(f);
	printf("\n");

	/* create z_2^n */
	z2n = mplf_z2n_create(f);

	mplf_z2n_set_reducer_preinv(z2n);

	mploc_z2n_init(t0, z2n);
	mploc_z2n_init(t1, z2n);

	push_time();

	/* REAL STUFF IS HERE !!!! */
	{
		mploc_z2n_t a6, tr;
		mploc_z2n_init(a6, z2n);
		mploc_z2n_init(tr, z2n);

		if (deg==DEFAULT_DEG) {
			int i;
			mploc_z2n_set_ui(a6,0,1);
			for(i=0;i<sizeof(a6_coeffs)/sizeof(a6_coeffs[0]);i++) {
				mploc_z2x_set_coeff_ui(a6,a6_coeffs[i],1);
			}
		} else {
			mploc_z2n_set_random(a6, prng, 1);
		}

		printf("a6=");
		mploc_z2n_print(a6);
		printf("\n");

		AGM(a6);

		BivariateAGM(tr, a6);
		/*
		   mploc_z2n_print(tr);
		   printf("\n");
		 */
		printTraceInHW(tr);
		printf("\n");

		mploc_z2n_clear(a6);
		mploc_z2n_clear(tr);
	}

	timings_set_dispmode(TIMINGS_ARE_SECONDS);
	pop_time_colon("Time in seconds");

	/* With default random source, I get:
defpol=1 + X^3 + X^6 + X^7 + X^163 + O(2^100)
a6=1 + X^3 + X^5 + X^8 + X^9 + X^10 + X^12 + X^13 + X^15 + X^16 + X^17 + X^18 + X^19 + X^20 + X^21 + X^22 + X^25 + X^29 + X^33 + X^36 + X^38 + X^39 + X^40 + X^41 + X^44 + X^47 + X^48 + X^49 + X^51 + X^55 + X^57 + X^60 + X^61 + X^62 + X^63 + X^64 + X^67 + X^70 + X^71 + X^72 + X^73 + X^74 + X^77 + X^78 + X^79 + X^84 + X^87 + X^88 + X^89 + X^90 + X^91 + X^94 + X^100 + X^102 + X^103 + X^104 + X^106 + X^107 + X^108 + X^111 + X^115 + X^117 + X^124 + X^128 + X^129 + X^130 + X^131 + X^132 + X^134 + X^136 + X^141 + X^142 + X^145 + X^146 + X^150 + X^154 + X^155 + X^158 + X^161 + O(2^1)
161088514449475591064501
	*/

	REALLOC_INFO();

	/* free everything */
	mploc_z2n_clear(t0);
	mploc_z2n_clear(t1);
	mplf_z2n_destroy(&z2n);
	mploc_z2x_clear(f);
	mplf_z2_destroy(&z2);
	return 0;
}
