/* $Id: teich2.c 169 2004-09-30 14:39:30Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>

#include <assert.h>

#include "manu.h"

#include "mploc.h"
#include "z2.h"
#include "z2x.h"
#include "z2n.h"

/* The p-adic valuation is given by mploc_z2x_valuation */

void mploc_z2x_EvenPart(mploc_z2x_ptr res, mploc_z2x_srcptr P)
{
	mp_size_t i;
	mploc_t coeff_aux;
	mplf_t field;

	field = P->coeff_field;
	mploc_z2_init(coeff_aux, field);

	for (i = 0; 2 * i <= mploc_z2x_degree(P); i++) {
		mploc_z2x_get_coeff(coeff_aux, P, 2 * i);
		mploc_z2x_set_coeff(res, i, coeff_aux);
	}

	mploc_z2x_normalize(res);

	mploc_z2_clear(coeff_aux);
}

void mploc_z2x_OddPart(mploc_z2x_ptr res, mploc_z2x_srcptr P)
{
	mp_size_t i;
	mploc_t coeff_aux;
	mplf_t field;

	field = P->coeff_field;
	mploc_z2_init(coeff_aux, field);

	for (i = 0; 2 * i + 1 <= mploc_z2x_degree(P); i++) {
		mploc_z2x_get_coeff(coeff_aux, P, 2 * i + 1);
		mploc_z2x_set_coeff(res, i, coeff_aux);
	}

	mploc_z2x_normalize(res);

	mploc_z2_clear(coeff_aux);
}

/* forward-define this */
static void mploc_z2x_canonical_lift_inc2(
		mploc_z2x_ptr res, mploc_z2x_srcptr f0,
		mploc_z2x_srcptr f1, mploc_z2x_srcptr V,
		mp_size_t N);

/*****************************************************************************/
/*                      Canonical lift -- recursive way                      */
/*****************************************************************************/

/* mploc_z2x_canonical_lift: The Teichmueller lift of polynomial fbar        */
/* Input: fbar (precision must be 1), N the precision                        */
/* Output: in the mploc_z2x_t res, stores the teichmueller lift to the       */
/*         precision N                                                       */

/* Note ; this function mistakenly computes (-1)^deg(f)f
 */
static void mploc_z2x_canonical_lift_pre(
		mploc_z2x_ptr res,
		mploc_z2x_srcptr fbar,
		mp_size_t N)
{
	/* f, Nprime, fprime, f0, f1, x, V */
	mp_size_t Nprime;
	mplf_t field;
	mploc_z2x_t fprime, f0, f1, aux, V, delta;

	assert(fbar->prec == 1);

	if (N == 1) {
		mploc_z2x_set(res, fbar);
		mploc_z2x_normalize(res);
		return;
	}

	// Warning: this line is wrong!!!  (N/2 is an integer...)
	//     Nprime = (mp_size_t) ceilf(N/2);
	// Better use iceildiv(x,y) from manu.h  
	Nprime = iceildiv(N, 2);
	mploc_z2x_force_precision(res, N);

	field = fbar->coeff_field;
	mploc_z2x_init(fprime, field);
	mploc_z2x_init(f0, field);
	mploc_z2x_init(f1, field);
	mploc_z2x_init(aux, field);
	mploc_z2x_init(V, field);
	mploc_z2x_init(delta, field);

	mploc_z2x_set_ui(fprime, 0, N);
	mploc_z2x_set_ui(f0, 0, N);
	mploc_z2x_set_ui(f1, 0, N);
	mploc_z2x_set_ui(aux, 0, N);
	mploc_z2x_set_ui(V, 0, N);
	mploc_z2x_set_ui(delta, 0, N);

	mploc_z2x_canonical_lift_pre(fprime, fbar, Nprime);
	mploc_z2x_force_precision(fprime, N);
	mploc_z2x_EvenPart(f0, fprime);
	mploc_z2x_OddPart(f1, fprime);

	/* V := modulo((fprime - f0^2 + x*f1^2)/2^Nprime, 2^(N - Nprime)); */
	mploc_z2x_sqr(aux, f0);
	mploc_z2x_sqr(V, f1);
	mploc_z2x_mul_xi(V, V, 1);
	mploc_z2x_sub(V, V, aux);
	mploc_z2x_add(V, V, fprime);

	assert(mploc_z2x_valuation(V) >= Nprime);

	mploc_z2x_rshift(V, V, Nprime);

	mploc_z2x_force_precision(f0, N - Nprime);
	mploc_z2x_force_precision(f1, N - Nprime);
	mploc_z2x_canonical_lift_inc2(delta, f0, f1, V, N - Nprime);
	mploc_z2x_lshift(res, delta, Nprime);
	mploc_z2x_add(res, res, fprime);

	mploc_z2x_clear(fprime);
	mploc_z2x_clear(f0);
	mploc_z2x_clear(f1);
	mploc_z2x_clear(aux);
	mploc_z2x_clear(V);
	mploc_z2x_clear(delta);
}

void mploc_z2x_canonical_lift(mploc_z2x_ptr res, mploc_z2x_srcptr fbar,
			      mp_size_t N)
{
	mploc_z2x_canonical_lift_pre(res,fbar,N);
	if (fbar->deg&1) mploc_z2x_neg(res,res);
}


/* Polynomials f0, f1 and V must be at precision N */
void mploc_z2x_canonical_lift_inc2(mploc_z2x_ptr res, mploc_z2x_srcptr f0,
				   mploc_z2x_srcptr f1, mploc_z2x_srcptr V,
				   mp_size_t N)
{
	mp_size_t Nprime;
	mploc_z2x_t deltaprime, delta0, delta1, Vprime, Delta, aux,
		    faux0, faux1, Vaux;
	mplf_t field;

	assert(f0->prec == N);
	assert(f1->prec == N);
	assert(V->prec == N);

	if (N == 1) {
		mploc_z2x_neg(res, V);
		mploc_z2x_normalize(res);
		return;
	}

	/* delta, Nprime, deltaprime, delta0, delta1, Vprime, Delta, x */
	// Same as above
	// Nprime = (mp_size_t) ceilf(N/2);
	Nprime = iceildiv(N, 2);
	mploc_z2x_force_precision(res, N);

	field = f0->coeff_field;
	mploc_z2x_init(deltaprime, field);
	mploc_z2x_init(delta0, field);
	mploc_z2x_init(delta1, field);
	mploc_z2x_init(Vprime, field);
	mploc_z2x_init(Delta, field);
	mploc_z2x_init(aux, field);
	mploc_z2x_init(faux0, field);
	mploc_z2x_init(faux1, field);
	mploc_z2x_init(Vaux, field);

	mploc_z2x_set_ui(deltaprime, 0, N);
	mploc_z2x_set_ui(delta0, 0, N);
	mploc_z2x_set_ui(delta1, 0, N);
	mploc_z2x_set_ui(Vprime, 0, N);
	mploc_z2x_set_ui(Delta, 0, N);
	mploc_z2x_set_ui(aux, 0, N);

	/* faux0, faux1, Vaux : f0, f1, V with precision Nprime */
	mploc_z2x_set_change_precision(faux0, f0, Nprime);
	mploc_z2x_set_change_precision(faux1, f1, Nprime);
	mploc_z2x_set_change_precision(Vaux, V, Nprime);

	mploc_z2x_canonical_lift_inc2(deltaprime, faux0, faux1, Vaux,
			Nprime);
	mploc_z2x_force_precision(deltaprime, N);
	mploc_z2x_EvenPart(delta0, deltaprime);
	mploc_z2x_OddPart(delta1, deltaprime);

	/* Vprime := modulo((V + deltaprime - 2*(f0*delta0 - */
	/* x*f1*delta1))/2^Nprime, 2^(N - Nprime)); */
	mploc_z2x_mul(aux, f0, delta0);
	mploc_z2x_mul(Vprime, f1, delta1);
	mploc_z2x_mul_xi(Vprime, Vprime, 1);
	mploc_z2x_sub(Vprime, aux, Vprime);
	mploc_z2x_lshift(Vprime, Vprime, 1);
	mploc_z2x_neg(Vprime, Vprime);
	mploc_z2x_add(Vprime, Vprime, deltaprime);
	mploc_z2x_add(Vprime, Vprime, V);

	assert(mploc_z2x_valuation(Vprime) >= Nprime);

	mploc_z2x_rshift(Vprime, Vprime, Nprime);

	/* faux0, faux1 : f0, f1 with precision N - Nprime */
	mploc_z2x_set_change_precision(faux0, f0, N - Nprime);
	mploc_z2x_set_change_precision(faux1, f1, N - Nprime);
	mploc_z2x_canonical_lift_inc2(Delta, faux0, faux1, Vprime,
			N - Nprime);

	/* delta := 2^Nprime*Delta */
	mploc_z2x_lshift(res, Delta, Nprime);
	mploc_z2x_add(res, res, deltaprime);

	mploc_z2x_clear(deltaprime);
	mploc_z2x_clear(delta0);
	mploc_z2x_clear(delta1);
	mploc_z2x_clear(Vprime);
	mploc_z2x_clear(Delta);
	mploc_z2x_clear(aux);
	mploc_z2x_clear(faux0);
	mploc_z2x_clear(faux1);
	mploc_z2x_clear(Vaux);

	mploc_z2x_normalize(res);
}
