#ifndef FROMKERNEL_H_
#define FROMKERNEL_H_

/* This file contains all asm inlines which can be produced from the
 * kernel directives. Depending on the functions defined in the kernel,
 * some will be activated, some won't
 */

#if	defined(KERNEL_HAS_UMUL_PPMM)
ASM_FUNCTION(SQR1c,__k0)
INLINER int SQR1c__k0(mp_limb_t * t,const mp_limb_t * s)
{
	int carryout;
	umul_ppmm(carryout,t[0],s[0],s[0]);
	return carryout;
}
#endif


#if	defined(KERNEL_HAS_UMUL_PPMM)
ASM_FUNCTION(MUL2,__k0)
INLINER void MUL2__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t hi;
	MUL_NO_OVERLAP();
	umul_ppmm(hi,r[0],s1[0],s2[0]);
	hi += s1[1]*s2[0];
	hi += s1[0]*s2[1];
	r[1]= hi;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM) &&	\
	defined(KERNEL_HAS_CTERM_2)
ASM_FUNCTION(ADDMUL2,__k0)
INLINER void ADDMUL2__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t hi,lo;

	MUL_NO_OVERLAP();

	umul_ppmm(hi,lo,s1[0],s2[0]);
	carry_terminate_2(hi,r[0],lo);
	hi   += s1[1]*s2[0];
	hi   += s1[0]*s2[1];
	r[1] += hi;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)
/* It seems that some sad effects tend to make this perform poorly
 * (slower than MUL2, which is unexpected, to say the least).
 */
ASM_FUNCTION(SQR2,__k0)
INLINER void SQR2__k0(mp_limb_t * t,const mp_limb_t * s)
{
	mp_limb_t p;
	SQR_NO_OVERLAP();
	umul_ppmm(t[1],t[0],s[0],s[0]);
	p=s[0]*s[1];
	p=p+p;
	t[1]+=p;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM) &&	\
	defined(KERNEL_HAS_CTERM_2)
INLINER int SQR2c__k0(mp_limb_t * t,const mp_limb_t * s)
{
	mp_limb_t u,v,w;

	SQR_NO_OVERLAP();

	umul_ppmm(u,t[0],s[0],s[0]);
	w=s[1];
	v=s[0]*w;
	carry_terminate_2(w,v,v);	// better by an inch ?!
	carry_terminate_2(w,u,v);
	t[1]= u;
	return w;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM) &&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)
/* This version in C yields 29.75 cycles with -O4 ; twice better than
 * Frederik's asm code (which suffers from a 15cycles penalty for
 * function call...
 */
ASM_FUNCTION(MUL3,__k0)
INLINER void MUL3__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t u,v,w,z;
	MUL_NO_OVERLAP();
	umul_ppmm(u,r[0],s1[0],s2[0]);
	umul_ppmm_and_carry(w,v,s1[1],s2[0],u);

	umul_ppmm_and_carry(z,u,s1[0],s2[1],v);
	r[1]=u;
	w+=s1[0]*s2[2];
	w+=s1[1]*s2[1];
	w+=s1[2]*s2[0];
	r[2]=w+z;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)
ASM_FUNCTION(ADDMUL3,__k0)
INLINER void ADDMUL3__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t u,v,w,z;
	MUL_NO_OVERLAP();
	umul_ppmm(u,z,s1[0],s2[0]);
	carry_terminate_2(u,r[0],z);
	umul_ppmm_and_carry(w,v,s1[1],s2[0],u);

	umul_ppmm_and_carry(z,u,s1[0],s2[1],v);
	carry_terminate_2(w,r[1],u);
	w+=s1[0]*s2[2];
	w+=s1[1]*s2[1];
	w+=s1[2]*s2[0];
	r[2]+=w+z;
}
#endif

#if	defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_CTERM_2)	&&	\
	defined(KERNEL_HAS_DBL2_TERM)
ASM_FUNCTION(SQR3,__k0)
INLINER void SQR3__k0(mp_limb_t * t,const mp_limb_t * s)
{
	mp_limb_t u,v,w;
	SQR_NO_OVERLAP();
	/* FIXME ; was there a bug here ? */
	umul_ppmm(u,t[0],s[0],s[0]);
	umul_ppmm(w,v,s[1],s[0]);
	double2_terminate(w,v);
	carry_terminate_2(w,u,v);
	t[1]=u;
	v=s[0]*s[2];
	w+=v;
	w+=v;
	w+=s[1]*s[1];
	t[2]=w;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)
ASM_FUNCTION(SQR3c,__k0)
INLINER int SQR3c__k0(mp_limb_t * t,const mp_limb_t * s)
{
	mp_limb_t u,v,w,z;
	SQR_NO_OVERLAP();
	umul_ppmm(u,t[0],s[0],s[0]);
	umul_ppmm_and_carry(w,v,s[1],s[0],u);

	umul_ppmm_and_carry(z,u,s[0],s[1],v);
	t[1]=u;
	
	umul_ppmm(u,v,s[1],s[1]);
	carry_terminate_2(u,w,z);
	carry_terminate_2(u,w,v);
	z=s[0]*s[2];
	carry_terminate_2(u,z,z);
	carry_terminate_2(u,w,z);
	t[2]=w;
	return u;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)
/* 53.71 cycles for this C routine !  Better than any asm sibling so far... */
ASM_FUNCTION(MUL4,__k0)
INLINER void MUL4__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t u,v,w,z;
	MUL_NO_OVERLAP();
	umul_ppmm(u,r[0],s1[0],s2[0]);
	umul_ppmm_and_carry(w,v,s1[1],s2[0],u);
	umul_ppmm_and_carry(z,u,s1[0],s2[1],v);
	r[1]=u;
	umul_ppmm_and_carry(v,u,s1[0],s2[2],w);
	carry_terminate_2(v,u,z);
	umul_ppmm(z,w,s1[1],s2[1]);
	carry_terminate_2(v,u,w);
	v+=z;
	umul_ppmm(z,w,s1[2],s2[0]);
	carry_terminate_2(v,u,w);
	r[2]=u;
	v+=z;
	v+=s1[3]*s2[0];
	v+=s1[2]*s2[1];
	v+=s1[1]*s2[2];
	r[3]=v+s1[0]*s2[3];
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)
ASM_FUNCTION(ADDMUL4,__k0)
INLINER void ADDMUL4__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t u,v,w,z;
	MUL_NO_OVERLAP();
	umul_ppmm(u,z,s1[0],s2[0]);
	carry_terminate_2(u,r[0],z);
	umul_ppmm_and_carry(w,v,s1[1],s2[0],u);
	umul_ppmm_and_carry(z,u,s1[0],s2[1],v);
	carry_terminate_2(z,r[1],u);
	umul_ppmm_and_carry(v,u,s1[0],s2[2],w);
	carry_terminate_2(v,u,z);
	umul_ppmm(z,w,s1[1],s2[1]);
	carry_terminate_2(v,u,w);
	v+=z;
	umul_ppmm(z,w,s1[2],s2[0]);
	carry_terminate_2(v,u,w);
	carry_terminate_2(v,r[2],u);
	v+=z;
	v+=s1[3]*s2[0];
	v+=s1[2]*s2[1];
	v+=s1[1]*s2[2];
	r[3]+=v+s1[0]*s2[3];
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)	&&	\
	defined(KERNEL_HAS_CTERM_3)	&&	\
	defined(KERNEL_HAS_DBL3_TERM)
//
// p <= (w-1)^2 = w^2-2w+1
// 2p <= 2w^2-4w+2
// x <= w-1
//
// 2p+x <= 2w^2-3w+1
//
// Peut-on avoir w^2 > 2p > w^2-w+1 = (w-1)^2-w ?
//
// w=10 w^2-w+1 = 91
//
// p = 8 * 6 ; 2p = 96 ; x=7 ; 2p+x = 103
// p = 8 * 7 ; 2p = 112; x=9 ; 2p+x = 121
// 
ASM_FUNCTION(SQR4,__k0)
INLINER void SQR4__k0(mp_limb_t * t,const mp_limb_t * s)
{
	mp_limb_t u,v,w,z;
	SQR_NO_OVERLAP();
	// all double products
	umul_ppmm(v,u,s[0],s[1]);
	umul_ppmm_and_carry(z,w,s[0],s[2],v);
	z+=s[1]*s[2];
	z+=s[0]*s[3];
	double3_terminate(z,w,u);

	umul_ppmm(v,t[0],s[0],s[0]);
	// Maybe one can prove that depth=2 is enough. But _very_
	// surprisingly, putting _2 has a net cost of 5.25 cycles !
	carry_terminate_3(z,w,v,u);
	t[1]=v;
	umul_ppmm(v,u,s[1],s[1]);
	carry_terminate_2(z,w,u);
	t[2]=w;
	t[3]=z+v;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)	&&	\
	defined(KERNEL_HAS_CTERM_3)	&&	\
	defined(KERNEL_HAS_CTERM_4)	&&	\
	defined(KERNEL_HAS_DBL3_PROP_ADD)
ASM_FUNCTION(SQR4c,__k0)
INLINER int SQR4c__k0(mp_limb_t * t,const mp_limb_t * s)
{
	mp_limb_t u,v,w,z,c;

	SQR_NO_OVERLAP();

	// all double products
	umul_ppmm(v,u,s[0],s[1]);
	umul_ppmm_and_carry(z,w,s[0],s[2],v);
	z+=s[0]*s[3];
	z+=s[1]*s[2];
	c=s[2];
	double3_propagate_add(c,z,w,u);

	umul_ppmm(v,t[0],s[0],s[0]);
	carry_terminate_4(c,z,w,v,u);
	t[1]=v;
	umul_ppmm(v,u,s[1],s[1]);
	carry_terminate_3(c,z,w,u);
	t[2]=w;
	carry_terminate_2(c,z,v);
	t[3]=z;

	return c;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)	&&	\
	defined(KERNEL_HAS_ADD2_STREAM)
/* This code is slightly slower than the asm version. The performance
 * depends largely on the CFLAGS:
 *
 * -O3 -fssa	 94.30 cycles
 * -O3		108.94 cycles	(-O2 -frename-registers is similar)
 * -O2		113.52 cycles
 *
 * We've got an example where appropriate CFLAGS yield a 17% speedup. Not
 * bad :-)
 *
 * WARNING : -fssa is tagged experimental, and indeed some suboptions do
 * mess up with the code badly (-fssa-ccp notably. -fssa-dce hasn't
 * wreaked havoc now, but ``dead code'' is hopefully absent in here
 * anyway).
 *
 * Surprisingly, the timings above seem to have disappeared :-( ; with
 * the said compilation flags, I get 109 instead of 94 ; additional
 * *fight* with gcc's optimizer gains 9 cycles, so I'm now at 100.00
 * cycles with -fomit-frame-pointer on top of the rest (and horrendous
 * 110 cycles with the very same code. Argh)
 */

/* as far as I can tell, this piece of code really mimicks the asm
 * version in asm/x86/x86raw.h . So it should be worthwhile to track down
 * as low as possible the difference between the two, using partial
 * expansions and timings. 15% difference seems to be a lot. Perhaps more
 * care on the asm constraints could be enough to grab back the
 * bottleneck.
 */

/* gcc is better than humans at tracking variable flow. Following this
 * constation, we keep a uniform association of the variable names
 * with the meaning.
 *
 * This is probably the reason why -fssa performs so good.
 */
ASM_FUNCTION(MUL5,__k0)
INLINER void MUL5__k0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t u,v,w,z;
	
	MUL_NO_OVERLAP();

	/* First, handle mul1 by s2[0] */
	z=s2[0];

	umul_ppmm(u,r[0],s1[0],z);
	/* some stuff pending in u */

	umul_ppmm_and_carry(w,v,s1[1],z,u);
	r[1]=v;
	u=w;
	/* some stuff pending in u */

	umul_ppmm_and_carry(w,v,s1[2],z,u);
	r[2]=v;
	u=w;
	/* some stuff pending in u */

	umul_ppmm_and_carry(w,v,s1[3],z,u);
	r[3]=v;
	u=w;
	/* some stuff pending in u */

	r[4]=u+s1[4]*z;
	/* Done with s2[0] */

	/**************************************/

	umul_ppmm(v,u,s1[0],s2[1]);
	/* some stuff pending in v,u ; w,z available */

	umul_ppmm(z,w,s1[1],s2[1]);
	add2_stream(v,r[1],z,w,u);
	/* some stuff pending in z,v ; w,u available */
	u=v;
	v=z;
	/* some stuff pending in v,u ; w,z available */
	
	umul_ppmm(z,w,s1[2],s2[1]);
	add2_stream(v,r[2],z,w,u);
	u=v;
	v=z;
	
	w=s1[3]*s2[1];
	carry_terminate_2(w,r[3],u);
	r[4]+=w+v;

	/**************************************/
	umul_ppmm(v,u,s1[0],s2[2]);

	umul_ppmm(z,w,s1[1],s2[2]);
	add2_stream(v,r[2],z,w,u);
	u=v;
	v=z;
	
	v+=s1[2]*s2[2];
	carry_terminate_2(v,r[3],u);
	r[4]+=v;

	/**************************************/
	
	/* wins a bit (2.5 cycles) this way when omit-fp is unset, and
	 * loses 1 cycle when set.
	 */
	umul_ppmm(v,u,s1[0],s2[3]);
	w=s1[0]*s2[4]+s1[1]*s2[3];
	carry_terminate_2(v,r[3],u);
	r[4]+=w+v;
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)	&&	\
	defined(KERNEL_HAS_ADD2_STREAM)
/* We want to check whether gcc with loop unrolling does a better job
 * than us... Apparently not...
 *
 * This peaks at 140cycles with -O3 -funroll-loops -fssa
 * (weird ; I thought -O3 had -funroll-loops anyway...)
 * -funroll-all-loops does no good.
 *
 * The problem is that -funroll-loops apparently cannot unroll these two
 * imbricated interdependent loops.
 *
 * At least that's still *much* better than mpn_mul_n with a funcall
 * (best cutoff is apparently at 3, with sth like 230 cycles)
 */
ASM_FUNCTION(MUL5,__k3)
INLINER void MUL5__k3(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	register mp_limb_t u,v,w,z;

	int i,j;
	const int n=5;
	
	MUL_NO_OVERLAP();

	/* Fill in the blanks first... */

	umul_ppmm(u,r[0],s1[0],s2[0]);
	/* some stuff pending in u */

	for(i=1;i<=n-2;i++) {
		umul_ppmm_and_carry(w,v,s1[i],s2[0],u);
		r[i]=v;
		u=w;
		/* some stuff pending in u */
		/* r[0..i] have been written */
	}

	r[n-1]=u+s1[n-1]*s2[0];

	/**************************************/

	for(j=1;j<=n-2;j++) {
		umul_ppmm(v,u,s1[0],s2[j]);
		/* some stuff pending in v,u ; w,z available */
		for(i=1;i+j-1<n-2;i++) {
			umul_ppmm(z,w,s1[i],s2[j]);
			add2_stream(v,r[i+j-1],z,w,u);
			/* add u to r[i+j-1] (was pending). Add w and the
			 * carry out to v. Add the carry out to z.
			 */
			u=v;
			v=z;
			/* some stuff pending in v,u ; w,z available.
			 * Add-up to r[i+j] is pending (in u) */

			/* Writes in r[j..j+i-1] have been performed */
		}

		ASSERT(j+i-1 == n-2);
		ASSERT(j+i == n-1);

		/* write to r[n-2] is pending, data is in u. */
	
		w=s1[i]*s2[j];
		carry_terminate_2(w,r[n-2],u);
		r[n-1]+=w+v;
	}

	r[n-1]+=s1[0]*s2[n-1];
}
#endif

#if     defined(KERNEL_HAS_UMUL_PPMM)	&&	\
	defined(KERNEL_HAS_UMUL_PPMM_C)	&&	\
	defined(KERNEL_HAS_CTERM_2)	&&	\
	defined(KERNEL_HAS_ADD2_STREAM)
/* fully expanded, this reaches 104 cycles. hence a 25% gain. */
ASM_FUNCTION(MUL5,__k4)
INLINER void MUL5__k4(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	mp_limb_t u,v,w,z;

	MUL_NO_OVERLAP();

	/* Fill in the blanks first... */
#define	N	5

	umul_ppmm(u,r[0],s1[0],s2[0]);
	/* some stuff pending in u */

#define	FIRSTLOOP(i)	\
		umul_ppmm(w,v,s1[i],s2[0]);	\
		carry_terminate_2(w,v,u);	\
		r[i]=v;				\
		u=w;
	
	FIRSTLOOP(1)
	FIRSTLOOP(2)
	FIRSTLOOP(3)

	r[N-1]=u+s1[N-1]*s2[0];

	/**************************************/

#define	INNERLOOP(i,j)	\
		umul_ppmm(z,w,s1[i],s2[j]);	\
		add2_stream(v,r[i+j-1],z,w,u);	\
		u=v;				\
		v=z;				\

#define	LOOPHEADER(j)	\
		umul_ppmm(v,u,s1[0],s2[j]);

#define	LOOPTRAILER(j)	\
		w=s1[N-1-j]*s2[j];			\
		carry_terminate_2(w,r[N-2],u);

	LOOPHEADER(1)
	INNERLOOP(1,1)
	INNERLOOP(2,1)
	LOOPTRAILER(1)
	r[N-1]+=w+v;

	LOOPHEADER(2)
	INNERLOOP(1,2)
	LOOPTRAILER(2)
	r[N-1]+=w+v;

	LOOPHEADER(3)
	LOOPTRAILER(3)
	r[N-1]+=w+v+s1[0]*s2[N-1];
#undef N
#undef FIRSTLOOP
#undef INNERLOOP
#undef LOOPHEADER
#undef LOOPTRAILER
}
#endif

#endif	/* FROMKERNEL_H_ */
