#ifndef X86RAW_H_
#define X86RAW_H_

#define MUL_4_BY_0      "0"
#define MUL_4_BY_1      "4"
#define MUL_4_BY_2      "8"
#define MUL_4_BY_3      "12"
#define MUL_4_BY_4      "16"
#define MUL_4_BY_5      "20"
#define MUL_4_BY_6      "24"
#define MUL_4_BY_7      "28"

#define CN(x,n,c)       [x] # c (x)
#define A(i)    MUL_4_BY_ ## i "(%[s1])"
#define B(i)    MUL_4_BY_ ## i "(%[s2])"
#define R(i)    MUL_4_BY_ ## i "(%%ebp)"
#define	S(i)	MUL_4_BY_ ## i "(%[s])"
#define	T(i)	MUL_4_BY_ ## i "(%[t])"

ASM_FUNCTION(ADDMUL1,__x86_1)
INLINER void ADDMUL1__x86_1(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	asm(	"movl	%2, %%eax\n\t"
		"imull	%1, %%eax\n\t"
		"addl	%%eax, %0\n"
		: "+r" (r[0]): "rm" (s1[0]), "rm" (s2[0]) : "eax" );
}

ASM_FUNCTION(SQR2,__x86_1)
INLINER void SQR2__x86_1(mp_limb_t * t,const mp_limb_t * s)
{
	asm volatile(
		"movl	"S(0)", %%eax	\n"
		"movl	%%eax,%%ebx	\n"
		"mull	%%eax		\n"
		"movl	%%eax,"T(0)"	\n"
		"movl	%%ebx,%%eax	\n"
		"movl	%%edx,%%ebx	\n"
		"movl	"S(1)",%%edx	\n"
		"mull	%%edx		\n"
		"addl	%%eax,%%ebx	\n"
		"addl	%%eax,%%ebx	\n"
		"movl	%%ebx,"T(1)"	\n"
		: : CN(t,0,r), CN(s,1,r): "eax","ebx","edx");
}

#if 0
/* There's apparently a bug with this one -- investigate [20060929]
 */
ASM_FUNCTION(SQR2c,__x86_1)
INLINER int SQR2c__x86_1(mp_limb_t * t,const mp_limb_t * s)
{
	register mp_limb_t p;
	register int carryout;

	p=s[0]*s[1];

	carryout=s[1];

	asm volatile ("mull	%2\n" : "=a" (t[0]), "=d" (t[1]) : "a" (s[0]));

	asm volatile (
		"	addl	%0, %0\n"
		"	adcl	$0, %1\n"
		"	addl	%0, %2\n"
		"	adcl	$0, %1\n"
		: "+r" (p), "+r" (carryout), "=g"(t[1]));
	return carryout;
}

/* There's apparently a bug with this one -- investigate [20060929]
 */
ASM_FUNCTION(SQR2c,__x86_2)
INLINER int SQR2c__x86_2(mp_limb_t * t,const mp_limb_t * s)
{
	register mp_limb_t p;
	register int carryout;

	asm ("mull	%2\n" : "=a" (t[0]), "=d" (t[1]) : "a" (s[0]));

	p=s[0]*s[1];
	carryout=s[1];

	asm volatile(
		"movl	"S(0)", %%eax	\n"
		"movl	%%eax,%%ebx	\n"
		"mull	%%eax		\n"
		"movl	%%eax,"T(0)"	\n"
		"movl	%%ebx,%%eax	\n"
		"movl	%%edx,%%ebx	\n"
		"movl	"S(1)",%%ecx	\n"
		"mull	%%ecx		\n"
		"addl	%%eax,%%ebx	\n"
		"adcl	$0,%%ecx	\n"
		"addl	%%eax,%%ebx	\n"
		"adcl	$0,%%ecx	\n"
		"movl	%%ebx,"T(1)"	\n"
		: "=c"(carryout) : CN(t,1,D), CN(s,2,S): "eax","ebx","edx");
	return carryout;
}
#endif

ASM_FUNCTION(SQR3,__x86_1)
INLINER void SQR3__x86_1(mp_limb_t * t,const mp_limb_t * s)
{
	asm volatile(
	/* First put all relevant double products (to do only the
	 * doubling at the end). */
		"movl	"S(0)", %%ebx	\n"
		"movl	"S(1)", %%eax	\n"
		"mull	%%ebx		\n"
		"addl	%%eax,%%eax	\n"
		"adcl	%%edx,%%edx	\n"
		"movl	%%eax, "T(1)"	\n"
		"movl	%%edx, "T(2)"	\n"

		"movl	"S(2)", %%eax	\n"
		"mull	%%ebx		\n"
		"addl	%%eax,%%eax	\n"
		"addl	%%eax, "T(2)"	\n"
	/* Now do the squares... */
		"movl	%%ebx,%%eax	\n"
		"mull	%%ebx		\n"
		"movl	%%eax,"T(0)"	\n"
		"addl	%%edx,"T(1)"	\n"
		"movl	"T(2)",%%ebx	\n"
		"adcl	$0,%%ebx	\n"
		"movl	"S(1)",%%eax	\n"
		"mull	%%eax		\n"
		"addl	%%eax,%%ebx	\n"
		"movl	%%ebx,"T(2)"	\n"
		
		: : CN(t,0,r), CN(s,1,r):
		"eax","edx","ebx");
}
ASM_FUNCTION(SQR3c,__x86_1)
INLINER int SQR3c__x86_1(mp_limb_t * t,const mp_limb_t * s)
{
	int carryout;

	asm volatile(
	/* First put all relevant double products (to do only the
	 * doubling at the end). */
		"movl	"S(0)", %%ebx	\n"
		"movl	"S(1)", %%eax	\n"
		"mull	%%ebx		\n"
		"addl	%%eax,%%eax	\n"
		"adcl	%%edx,%%edx	\n"
		"movl	$0,%%ecx	\n"
		"movl	%%eax, "T(1)"	\n"
		"movl	%%edx, "T(2)"	\n"
		"adcl	$0,%%ecx	\n"

		"movl	"S(2)", %%eax	\n"
		"mull	%%ebx		\n"
		"addl	%%eax,%%eax	\n"
		"adcl	$0,%%ecx	\n"
		"addl	%%eax, "T(2)"	\n"
		"adcl	$0,%%ecx	\n"
	/* Now we have the current carry out in ecx ; do the squares... */
		"movl	%%ebx,%%eax	\n"
		"mull	%%ebx		\n"
		"movl	%%eax,"T(0)"	\n"
		"addl	%%edx,"T(1)"	\n"
		"movl	"T(2)",%%ebx	\n"
		"adcl	$0,%%ebx	\n"
		"adcl	$0,%%ecx	\n"
		"movl	"S(1)",%%eax	\n"
		"mull	%%eax		\n"
		"addl	%%eax,%%ebx	\n"
		"adcl	$0,%%ecx	\n"
		"movl	%%ebx,"T(2)"	\n"
		"adcl	%%edx,%%ecx	\n"
		
		: "=c"(carryout) : CN(t,1,D), CN(s,2,S):
		"eax","edx","ebx");

	return carryout;
}

ASM_FUNCTION(SQR4,__x86_1)
INLINER void SQR4__x86_1(mp_limb_t * t,const mp_limb_t * s)
{
	asm volatile(
	/* same scheme. double products first. */

		"movl	"S(0)", %%ebx	\n"

		"movl	"S(1)", %%eax	\n"
		"mull	%%ebx		\n"
		"movl	%%eax, "T(1)"	\n"
		"movl	%%edx, "T(2)"	\n"

		"movl	"S(3)", %%eax	\n"
		"mull	%%ebx		\n"
		"movl	%%eax, "T(3)"	\n"

		"movl	%%ebx,%%eax	\n"
	/* grab s2 in ebx */
		"movl	"S(2)", %%ebx	\n"
		"mull	%%ebx		\n"
		"addl	%%eax, "T(2)"	\n"
		"adcl	%%edx, "T(3)"	\n"

		"movl	"S(1)", %%eax	\n"
		"mull	%%ebx		\n"
		"addl	%%eax, "T(3)"	\n"
		
	/* shifts ; actually additions... */

		"movl	"T(1)",%%eax	\n"
		"movl	"T(2)",%%ebx	\n"
		"movl	"T(3)",%%edx	\n"
		"addl	%%eax, %%eax	\n"
		"adcl	%%ebx, %%ebx	\n"
		"adcl	%%edx, %%edx	\n"
		"movl	%%eax,"T(1)"	\n"
		"movl	%%ebx,"T(2)"	\n"
		"movl	%%edx,"T(3)"	\n"

		
	/* Now do the squares... */
		"movl	"S(0)",%%eax	\n"
		"xorl	%%ebx,%%ebx	\n"
		"mull	%%eax		\n"
		"movl	%%eax,"T(0)"	\n"
		"addl	%%edx,"T(1)"	\n"
		"adcl	$0,%%ebx	\n"
		"movl	"S(1)",%%eax	\n"
		"mull	%%eax		\n"
		"addl	%%ebx,%%eax	\n"
		"adcl	$0,%%edx	\n"	/* NO carry */
		"addl	%%eax,"T(2)"	\n"
		"adcl	%%edx,"T(3)"	\n"
		
		: : CN(t,0,r), CN(s,1,r):
		"eax","edx","ebx");
}

ASM_FUNCTION(SQR4c,__x86_1)
INLINER int SQR4c__x86_1(mp_limb_t * t,const mp_limb_t * s)
{
	int carryout;

	asm volatile(
		"movl	"S(0)", %%ebx	\n"

		"movl	"S(1)", %%eax	\n"
		"mull	%%ebx		\n"
		"movl	%%eax, "T(1)"	\n"
		"movl	%%edx, "T(2)"	\n"

		"movl	"S(3)", %%eax	\n"
		"mull	%%ebx		\n"
		"movl	%%eax, "T(3)"	\n"

		"movl	%%ebx,%%eax	\n"
	/* grab s2 in ecx : it impacts on the carry at the end */
		"movl	"S(2)", %%ecx	\n"
		"mull	%%ecx		\n"
		"addl	%%eax, "T(2)"	\n"
		"adcl	%%edx, "T(3)"	\n"

		"movl	"S(1)", %%eax	\n"
		"mull	%%ecx		\n"
		"addl	%%eax, "T(3)"	\n"
		
	/* shifts ; actually additions... */

		"movl	"T(1)",%%eax	\n"
		"movl	"T(2)",%%ebx	\n"
		"movl	"T(3)",%%edx	\n"
		"addl	%%eax, %%eax	\n"
		"adcl	%%ebx, %%ebx	\n"
		"adcl	%%edx, %%edx	\n"
		"movl	%%eax,"T(1)"	\n"
		"movl	%%ebx,"T(2)"	\n"
		"movl	%%edx,"T(3)"	\n"
		"adcl	$0,%%ecx	\n"

		
	/* Now we have the current carry out in ecx ; do the squares... */
		"movl	"S(0)",%%eax	\n"
		"xorl	%%ebx,%%ebx	\n"
		"mull	%%eax		\n"
		"movl	%%eax,"T(0)"	\n"
		"addl	%%edx,"T(1)"	\n"
		"adcl	$0,%%ebx	\n"
		"movl	"S(1)",%%eax	\n"
		"mull	%%eax		\n"
		"addl	%%ebx,%%eax	\n"
		"adcl	$0,%%edx	\n"	/* NO carry */
		"addl	%%eax,"T(2)"	\n"
		"adcl	%%edx,"T(3)"	\n"
		"adcl	$0,%%ecx	\n"
		
		: "=c"(carryout) : CN(t,1,D), CN(s,2,S):
		"eax","edx","ebx");

	return carryout;
}

#if 0
/* MUL5__x86_1 is broken, don't use it  */
ASM_FUNCTION(MUL5,__x86_1)
INLINER void MUL5__x86_1(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	volatile mp_limb_t a420,a212,a20,a424,a216,a24,a28;
	MUL_NO_OVERLAP();
	asm volatile(
"# This code duplicates the code generated by GCC for byhand_C\n\t"
//  This is the code generated by MUL5_byhand_C
//  There must be a reason why it is so significantly bad compared to the
//  handcrafted asm code.
	"movl	%[s2], %%ebx\n\t"
//"	movl	%[s1], %%edx\n\t"
	"movl	%[s1], %%esi\n\t"
	"movl	%[r], %%edi\n\t"
//  weird and certainly inefficient. movl	edx, esi must be better.
//  And anyway, the movl	into edx is dead code if we replace (%%edx)
//  by (%%esi) below
//  at this point, esi == edx == s1 , edi == r, ebx == s2
	"movl	(%%ebx), %%ecx\n\t"
	"movl	(%%esi), %%eax\n\t"
	"mull	%%ecx\n\t"
	"movl	%%eax, (%%edi)\n\t"
//  first umul is done.
//  get s1[1]
	"movl	4(%%esi), %%eax\n\t"
//  save out edx
	"movl	%%edx, %%ebx\n\t"
	"mull	%%ecx\n\t"
//   w,v is edx,eax
//   //  WEEIIIRD !!! Why not eax ?
//   "movl	%%eax, %%edi\n\t"
//   //  take r back. silly, since it was in edi already...
//   "movl	%[r], %%eax\n\t"
//   //  again, this is very probably useless.
//   "movl	%%edx, %%esi\n\t"
//   "addl	%%ebx, %%edi\n\t"
//   "adcl	$0, %%esi\n\t"
//   "movl	%%edi, 4(%%eax)\n\t"
//   //  OK, that's it ; the last six insns could have been replaced much more
//   //  thoughtfully by the following:
	"addl	%%ebx, %%eax\n\t"
	"adcl	$0, %%edx\n\t"
	"movl	%%eax, 4(%%edi)\n\t"
//  that's at the very least 3 insns cheaper. And it doesn't touch esi or
//  edi.
	"movl	8(%%esi), %%eax\n\t"
	"movl	%%edx, %%ebx\n\t"
	"mull	%%ecx\n\t"
	"addl	%%ebx, %%eax\n\t"
	"adcl	$0, %%edx\n\t"
	"movl	%%eax, 8(%%edi)\n\t"
//  Hmmm. Looks like we resort to temporary storage while it's not surely
//  needed.
	"movl	12(%%esi), %%eax\n\t"
	"movl	%%edx, %%ebx\n\t"
	"mull	%%ecx\n\t"
	"addl	%%ebx, %%eax\n\t"
	"adcl	$0, %%edx\n\t"
	"movl	%%eax, 12(%%edi)\n\t"
//  same deal.
	"movl	16(%%edi), %%eax\n\t"
	"imull	%%ecx, %%eax\n\t"
	"addl	%%edx, %%eax\n\t"
	"movl	%%eax, 16(%%edi)\n\t"

	//"movl	%%ebx, %[a212]\n\t"
	//"movl	%%ebx, 12(%%edx)\n\t"
	//"imull	16(%%ebx), %%ecx\n\t"
	//"addl	%%ecx, %%eax\n\t"
	//"movl	%%eax, 16(%%edx)\n\t"
//  first block (all multiplications by s2[0]) done.
	"movl	%[s2], %%ebx\n\t"
	"movl	(%%esi), %%eax\n\t"
	"movl	(%%ebx), %%ecx\n\t"

	"mull	%%ecx\n\t"

	"addl	%%eax, 4(%%edi)\n\t"
	"adcl	$0, %%edx\n\t"
	"movl	%%edx, %%ebx\n\t"
	"movl	4(%%esi), %%eax\n\t"

	"mull	%%ecx\n\t"
	"addl	%%eax, %%ebx\n\t"
	"adcl	$0, %%edx\n\t"
	"addl	%%eax, 8(%%edi)\n\t"
	"adcl	$0, %%edx\n\t"
	"movl	%%edx, %%ebx\n\t"
	"movl	8(%%esi), %%eax\n\t"

	"mull	%%ecx\n\t"
	"addl	%%eax, %%ebx\n\t"
	"adcl	$0, %%edx\n\t"
	"addl	%%eax, 12(%%edi)\n\t"
	"adcl	$0, %%edx\n\t"
	"movl	8(%%esi), %%eax\n\t"

	"imull	%%ecx,%%eax\n\t"
	"addl	%%edx,%%eax\n\t"
	"addl	%%eax,16(%%edi)\n\t"
	
	// Done third block -- much shorter !
	// but we only win marginally -- probably better could be
	// achieved.
	
	/*
	"movl	%%edx, %%ebx\n\t"
	"movl	%[s1], %%edx\n\t"
	"movl	%%eax, %%ecx\n\t"
	"movl	4(%%edx), %%eax\n\t"
	"movl	%[s2], %%edx\n\t"
	"mull	4(%%edx)\n\t"
	"movl	%%edx, %[a424]\n\t"
	"addl	%%ecx,%%edi\n\t"
	"adcl	%%eax,%%ebx\n\t"
	"adcl	$0,%%edx\n\t"
	"movl	%%edx, %%ecx\n\t"
	"movl	%[r], %%edx\n\t"
	"movl	%%edi, 4(%%edx)\n\t"
	"movl	%[s1], %%edi\n\t"
	"movl	8(%%edi), %%eax\n\t"
	"movl	%[s2], %%edi\n\t"
	"mull	4(%%edi)\n\t"
	"addl	%%ebx,%%esi\n\t"
	"adcl	%%eax,%%ecx\n\t"
	"adcl	$0,%%edx\n\t"
	"movl	%[r], %%eax\n\t"
	"movl	%%esi, %%ebx\n\t"
	"movl	%%esi, 8(%%eax)\n\t"
	"movl	%[s1], %%esi\n\t"
	"movl	4(%%edi), %%eax\n\t"
	"movl	12(%%esi), %%edi\n\t"
	"imull	%%edi, %%eax\n\t"
	"movl	%[a212], %%edi\n\t"
	"addl	%%ecx, %%edi\n\t"
	"adcl	$0, %%eax\n\t"
	"movl	%%edi, %[a216]\n\t"
	"addl	%%edx, %%eax\n\t"
	"movl	%[a20], %%edx\n\t"
	"movl	%[r], %%ecx\n\t"
	"addl	%%edx, %%eax\n\t"
	"movl	%%edi, 12(%%ecx)\n\t"
	"movl	%%eax, 16(%%ecx)\n\t"
	*/
//  OK. Done second block.
	"movl	(%%esi), %%edx\n\t"
	"movl	%[s2], %%ecx\n\t"
	"movl	%%edx, %%eax\n\t"
// APP
	"mull	8(%%ecx)\n\t"
// NO_APP
	"movl	%%eax, %%ecx\n\t"
	"movl	%%edx, %%edi\n\t"
	"movl	4(%%esi), %%eax\n\t"
	"movl	%[s2], %%esi\n\t"
// APP
	"mull	8(%%esi)\n\t"
	"	addl	%%ecx,%%ebx\n\t"
	"adcl	%%eax,%%edi\n\t"
	"adcl	$0,%%edx\n\t"
// NO_APP
	"movl	%%ebx, %%eax\n\t"
	"movl	%[r], %%ebx\n\t"
	"movl	%%edi, %%ecx\n\t"
	"movl	%%eax, 8(%%ebx)\n\t"
	"movl	%[s1], %%ebx\n\t"
	"movl	8(%%esi), %%eax\n\t"
	"movl	8(%%ebx), %%edi\n\t"
	"imull	%%edi, %%eax\n\t"
	"movl	%[a216], %%edi\n\t"
	"addl	%%eax, %%edx\n\t"
	"movl	%%edx, %%esi\n\t"
// APP
	"	addl	%%ecx, %%edi\n\t"
	"adcl	$0, %%esi\n\t"
// NO_APP
	"movl	%[r], %%ecx\n\t"
	"movl	%%edi, 12(%%ecx)\n\t"
	"movl	%[a24], %%eax\n\t"
	"addl	%%eax, %%esi\n\t"
	"movl	%%esi, 16(%%ecx)\n\t"
//  Done third.
	"movl	(%%ebx), %%ebx\n\t"
	"movl	%%esi, %[a28]\n\t"
	"movl	%[s2], %%esi\n\t"
	"movl	%%ebx, %%eax\n\t"
// APP
	"mull	12(%%esi)\n\t"
// NO_APP
	"movl	%%eax, %%ecx\n\t"
	"movl	16(%%esi), %%eax\n\t"
	"imull	%%eax, %%ebx\n\t"
	"movl	12(%%esi), %%eax\n\t"
	"movl	%[s1], %%esi\n\t"
	"imull	4(%%esi), %%eax\n\t"
	"addl	%%eax, %%ebx\n\t"
	"movl	%%edx, %%eax\n\t"
// APP
	"	addl	%%ecx, %%edi\n\t"
	"adcl	$0, %%eax\n\t"
// NO_APP
	"movl	%%edi, %%ecx\n\t"
	"movl	%[r], %%edi\n\t"
	"movl	%%ecx, 12(%%edi)\n\t"
	"leal	(%%eax,%%ebx), %%ecx\n\t"
	"movl	%[s2], %%eax\n\t"
	"movl	%[a28], %%ebx\n\t"
	"addl	%%ebx, %%ecx\n\t"
	"movl	%%ecx, 16(%%edi)\n"
//  And that's it. Afterwards come 15 other expansions of the same
//  sequence
: [a20]"=m"(a20), [a24]"=m"(a24), [a28]"=m"(a28), [a212]"=m"(a212), [a216]"=m"(a216), [a420]"=m"(a420), [a424]"=m"(a424): [r]"m"(r), [s1]"m"(s1), [s2]"m"(s2) : "eax", "ebx", "ecx", "edx", "esi", "edi", "memory");
}
#endif

#undef	MUL_4_BY_0
#undef	MUL_4_BY_1
#undef	MUL_4_BY_2
#undef	MUL_4_BY_3
#undef	MUL_4_BY_4
#undef	MUL_4_BY_5
#undef	MUL_4_BY_6
#undef	MUL_4_BY_7
#undef	A
#undef	B
#undef	R
#undef	S
#undef	T
#undef	CN

#endif	/* X86RAW_H_ */
