/* $Id: z2n_cxx.hpp 194 2004-10-13 15:36:16Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef MPLOC_Z2N_CXX_H_
#define MPLOC_Z2N_CXX_H_

// It *is* possible to do something generic with the _field interface,
// since much of it is shared ; however, things would not be a direct
// translation from C to C++, so we stick to a dumb approach.
// For the record, one could make the different instances of the field
// class inherit from the same generic parent class. The benefit would be
// shared code for most of the ctor / dtor stuff.

#ifndef	CTORLOG
#define	CTORLOG(x)	x
#endif

class Mploc_z2n_field {
    public:
	mplf_t		f;

	inline const mplf_t cval() const { return f; }
	
	inline Mploc_z2n_field(const z2_poly& p):f(mplf_z2n_create(p.cval())) {}
	inline Mploc_z2n_field(const mplf_t sf) : f(sf) {}
	inline ~Mploc_z2n_field() {
		mplf_z2n_destroy(&f);
	}

	inline Mploc_z2n_field(const Mploc_z2n_field &src) :
		f(mplf_z2n_create(src.def_poly().cval()))
	{
		CTORLOG(std::cerr  << "CALLING COPY !\n");
	}

	inline Mploc_z2n_field& operator=(const Mploc_z2n_field& s)
	{
		std::cerr  << "Doing field=field !\n";
		f=mplf_z2n_create(s.def_poly().cval());
		return *this;
	}

	inline z2_poly& def_poly() const {
		return * new z2_poly(mplf_z2n_def_poly(f));
	}
	inline void set_reducer(struct z2x_reducer_t * r) {
		mplf_z2n_set_reducer(f,r);
	}
	static void inversion_configure(int m, int s) {} ;
};

class z2n_elt {
    public:
	mploc_z2x_t	x;

	typedef Mploc_z2n_field	field;

	//class Mploc_z2n_field& f;	// a bit unfortunate.

	inline mploc_z2x_srcptr cval() const { return x; }
	inline mploc_z2x_ptr cval() { return x; }

	inline z2n_elt(const mplf_t f) {
		mploc_z2n_init(x, f);
	}
	inline z2n_elt(const Mploc_z2n_field& sf) {
		mploc_z2n_init(x,sf.cval());
	}
	inline z2n_elt(const z2n_elt& src) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
		mploc_z2n_init(x,src.cval()->f);
		mploc_z2n_set(x,src.cval());
	}
	inline ~z2n_elt() {
		mploc_z2n_clear(x);
	}
	inline void set_reduce(const z2n_elt& y) {mploc_z2n_reduce(x,y.cval()); }
	inline void reduce() { set_reduce(*this); }
	inline int is_reduced() { return mploc_z2n_is_reduced(x); }
	inline void normalize() { mploc_z2n_normalize(x); }

	inline void force_precision(mp_size_t k) {
		mploc_z2n_force_precision(x,k);
	}
	
	inline void set(const z2n_elt& y, mp_size_t k = -1) {
		if (k==-1) k=y.x->prec;
		mploc_z2n_set_change_precision(x,y.cval(),k);
	}
	inline void set(mp_limb_t u, mp_size_t k) {
		mploc_z2n_set_ui(x,u,k);
	}
	inline void set(mpz_t z, mp_size_t k) {
		mploc_z2n_set_mpz(x,z,k);
	}
	static inline void set(z2n_elt&  z, const z2n_elt&  y) {z.set(y);}

	inline void random(gmp_randstate_t prng, mp_size_t k) {
		mploc_z2n_set_random(x,prng,k);
	}
	static inline void random(z2n_elt&  z,gmp_randstate_t prng,mp_size_t k){
		z.random(prng,k);
	}
	inline void random(gmp_randstate_t prng, mp_size_t d, mp_size_t k) {
		random(prng,k);
	}
	static inline void random(z2n_elt&  z, gmp_randstate_t prng,
			mp_size_t d, mp_size_t k) {
		z.random(prng,d,k);
	}

	inline int ok() const {
		return mploc_z2n_is_good_looking(x);
	}

	inline int valuation() const {
		return mploc_z2n_valuation(x);
	}

	inline int operator==(const z2n_elt& y) const {
		return mploc_z2n_eq(x,y.cval());
	}
	inline int operator==(mp_limb_t y) const {
		return mploc_z2n_eq_ui(x,y);
	}
	/* Arithmetic operations are handled by static member functions.
	 * It's preferrable to a false illusion of a well-thought
	 * interface (don't have time for this :-))
	 */
	static inline void add(z2n_elt& z, const z2n_elt& x, const z2n_elt& y) {
		mploc_z2n_add(z.cval(),x.cval(),y.cval());
	}
	static inline void add(z2n_elt& z, const z2n_elt& x, mp_limb_t y) {
		mploc_z2n_add_ui(z.cval(),x.cval(),y);
	}

	static inline void sub(z2n_elt& z, const z2n_elt& x, const z2n_elt& y) {
		mploc_z2n_sub(z.cval(),x.cval(),y.cval());
	}
	static inline void sub(z2n_elt& z, const z2n_elt& x, mp_limb_t y) {
		mploc_z2n_sub_ui(z.cval(),x.cval(),y);
	}

	static inline void neg(z2n_elt& x, const z2n_elt& y) {
		mploc_z2n_neg(x.cval(),y.cval());
	}

	static inline void mul(z2n_elt& z, const z2n_elt& x, const z2n_elt& y) {
		mploc_z2n_mul(z.cval(),x.cval(),y.cval());
	}
	static inline void sqr(z2n_elt&  z, const z2n_elt&  x) {
		mploc_z2n_sqr(z.cval(),x.cval());
	}
	static inline void inv(z2n_elt& x, const z2n_elt& y) {
		mploc_z2n_inv(x.cval(),y.cval());
	}
	static inline int invsqrt(z2n_elt& x, const z2n_elt& y) {
		return mploc_z2n_invsqrt(x.cval(),y.cval());
	}
	static inline int sqrt(z2n_elt& x, const z2n_elt& y) {
		return mploc_z2n_sqrt(x.cval(),y.cval());
	}
	static inline void lshift(z2n_elt& z, const z2n_elt& x, mp_size_t k) {
		mploc_z2n_lshift(z.cval(),x.cval(),k);
	}
	static inline void rshift(z2n_elt& z, const z2n_elt& x, mp_size_t k) {
		mploc_z2n_rshift(z.cval(),x.cval(),k);
	}
};

inline std::ostream& operator<< (std::ostream& o, const z2n_elt& x) {
	/* FIXME : Finish with an o.write ! */
	mploc_z2n_print(x.cval());
	return o;
}

#endif	/* MPLOC_Z2N_CXX_H_ */
