/* $Id: field_def.h 91 2004-09-16 15:02:07Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef FIELD_DEF_H_
#define FIELD_DEF_H_

/*
 * Lazy reduction is allowed all around in the field manipulation routines.
 *
 * Call ->reduce to be certain of the uniqueness of your results.
 *
 * Equality tests always do it.
 */

struct field_aux_data;

/* At present, some methods might not always be provided. sqrt is one of
 * them (sone for prime fields)
 */

#define ELT	mp_limb_t *
#define cELT	const ELT
#define	K	const struct field *

struct field {
	void (*add)	(ELT, cELT, cELT,	K);
	void (*sub)	(ELT, cELT, cELT,	K);
	void (*neg)	(ELT, cELT,		K);
	void (*addto)	(ELT, cELT,		K);
	void (*mul)	(ELT, cELT, cELT,	K);
	void (*addmul)	(ELT, cELT, cELT,	K);
	void (*inv)	(ELT, cELT,		K);
	void (*set_int)	(ELT, const int,	K);
	void (*set_mpn)	(ELT, cELT, const mp_size_t,	K);
	void (*set_mpz)	(ELT, const mpz_t,	K);
	int  (*is_int)	(ELT, const int,K);
	void (*exp)	(ELT, cELT, cELT, const mp_size_t, K);
	int  (*eq)	(ELT, ELT,	K);
	void (*set)	(ELT, cELT,	K);
	void (*reduce)	(ELT,		K);
	void (*conjugate)(ELT, cELT,	K);
	int  (*sqrt)    (ELT, cELT,	K);
	void (*print)	(cELT,		K);
	void (*destroy)	(struct field *);
	struct field *	base_field;
	unsigned int degree;
	mp_limb_t *	modulus;	/* only meaningful for prime fields */
	mp_limb_t *	minpoly;	/* only meaningful for normal exts. */
	mp_size_t size;	
	struct field_aux_data *	aux_data;/* might be NULL */
};

#undef ELT
#undef cELT
#undef K


#endif	/* FIELD_DEF_H_ */
