/* $Id: rf_layer_gfpn_ntl.cpp 101 2004-09-17 15:28:03Z /CN=thome/emailAddress=thome@loria.fr $ */

#undef __DEPRECATED

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <gmp.h>
#include <gmp/config.h>
#include <gmp/gmp-impl.h>
#include <gmp/longlong.h>

#include <NTL/ZZ_pX.h>
#include <NTL/ZZ_pXFactoring.h>
#include <NTL/ZZ_pE.h>
#include <NTL/ZZ_pEX.h>
#include <NTL/ZZ_pEXFactoring.h>

#include <cassert>

extern "C" {
#include "mploc.h"
#include "manu.h"

#include "zp.h"
#include "zpx.h"
#include "zpn.h"
}

#include "rf_layer_gfpn_ntl.hpp"

NTL_USE_NNS

void mpz_to_ZZ(ZZ& c, mpz_t& x)
{
	ASSERT(SIZ(x)>=0);
	c=0;
	for(int i=SIZ(x)-1;i>=0;i--) {
		LeftShift(c,c,GMP_LIMB_BITS);
		c=c+to_ZZ(PTR(x)[i]);
	}
}

void ZZ_to_mpz(mpz_t& x, ZZ& arg)
{
	ZZ c(arg);

	mpz_realloc2(x,arg.size() * GMP_LIMB_BITS);
	mpz_set_ui(x,0);

	if (arg==0) { return; }

	int s=(c<0);
	if (s) c=-c;
	
	ZZ small,word;

	LeftShift(word,to_ZZ(1UL),GMP_LIMB_BITS);
	for(int i=0;c!=0;i++) {
		DivRem(c,small,c,word);
		ASSERT(i<ALLOC(x));
		PTR(x)[i]=to_ulong(small);
	}
		
	if (s) mpz_neg(x,x);
}

void ntl_attach_GFPE(mplf_t f)
{
	ZZ	c;
	mpz_to_ZZ(c,f->p);
	ZZ_p::init(c);

	ZZ_pX p;
	p=0;
	mploc_zpx_srcptr pptr=mplf_zpn_def_poly(f);
	mpz_t	t;

	mpz_init(t);
	for(int i=0;i<=pptr->deg;i++) {
		mpz_mod(t,pptr->c[i],f->p);
		mpz_to_ZZ(c,t);
		SetCoeff(p,i,to_ZZ_p(c));
	}
	mpz_clear(t);

	ZZ_pE::init(p);
}

void ntl_detach_GFPE(mplf_t f)
{
	/*
	ZZ_pX p;
	p=0;
	ZZ_pE::init(p);
	*/
	// We prefer to do nothing here. The above fails badly, as NTL
	// checks pretty early that it's been given something sane (and 0
	// isn't).
}

rf_elt new_rfe(mplf_t f UNUSED_VARIABLE)
{
	return (void*) new ZZ_pE();
}

void _delete_rfe(rf_elt * a, mplf_t f UNUSED_VARIABLE)
{
	delete (ZZ_pE *) *a;
	*a=NULL;
}

void rfe_set_mploc_coeff(rf_elt xbar, mploc_zpx_srcptr x, mp_size_t k, mplf_t f)
{
	ZZ_pX p;
	p=0;
	ASSERT(k<x->prec);
	mpz_t t;
	ZZ	c;

	mpz_init(t);
	for(int i=0;i<=x->deg;i++) {
		mpz_fdiv_q(t,x->c[i],p_to_the(k,
					mplf_zpn_def_poly(f)->coeff_field));
		mpz_mod(t,t,f->p);
		mpz_to_ZZ(c,t);
		SetCoeff(p,i,to_ZZ_p(c));
	}
	mpz_clear(t);
	*(ZZ_pE*)xbar=to_ZZ_pE(p);
}

void rfe_set_mploc(rf_elt xbar, mploc_zpx_srcptr x, mplf_t f)
{
	rfe_set_mploc_coeff(xbar,x,0,f);
}

void mploc_set_rfe(mploc_zpx_ptr x, rf_elt a, mplf_t f)
{
	const ZZ_pX& p=rep(*(ZZ_pE*)a);
	mploc_zpx_clobber(x);
	mploc_zpx_realloc(x,1,f->degree-1,1);
	x->deg=f->degree-1;
	ZZ t;
	for(int i=0;i<f->degree;i++) {
		t=rep(coeff(p,i));
		ZZ_to_mpz(x->c[i],t);
	}
	mploc_zpx_normalize(x);
}

/* This one is more subtle than mploc_set_rfe */
void mploc_coeff_set_rfe(mploc_zpx_ptr x, mp_size_t k, rf_elt a, mplf_t f)
{
	//const ZZ_pX& p=rep(*(ZZ_pE*)a);
	ASSERT(k<x->prec);

	/* Make sure there's room ! */
	mploc_zpx_realloc(x,1,f->degree-1,x->prec);

	x->deg=f->degree-1;
	for(int i=0;i<f->degree;i++) {
		BUG();
#if 0
		/* Finishh this */
		x->c[i*x->cw + wi] &= ~(1UL<<wp);
		x->c[i*x->cw + wi] |= (rep(coeff(p,i))&1UL)<<wp;
#endif
	}
	mploc_zpx_normalize(x);
}

#if 0
/* Useful for debugging */
#include <iostream>
#include <string>

static std::ostream& operator<<(std::ostream& o, const rf_elt& pp)
{
	int k=0;
	const ZZ_pE& p=*(ZZ_pE*)pp;
	//std::ostringstream o;
	for(int i=0;i<=deg(rep(p));i++) {
		if (coeff(rep(p),i)==0) continue;
		if (k) o << " + ";
		if (i==0) { o << coeff(rep(p),i); k++; continue; }
		if (coeff(rep(p),i)!=0) { o << coeff(rep(p),i) << "*"; }
		o << "X";
		if (i>1) o << "^" << i;
		k++;
	}
	if (!k) o << "0";
	return o;
}

void disp(const rf_elt& p)
{
	cout << p;
}
#endif

int RF_sqrt(rf_elt y,rf_elt x, mplf_t f)
{
	ZZ_pE&	xe=*(ZZ_pE*)x;
	ZZ_pE&	ye=*(ZZ_pE*)y;

	ZZ_pEX	p;
	// How should I compute a square root w/ NTL ???
	
	SetCoeff(p,2,1);
	SetCoeff(p,0,-xe);

	vec_pair_ZZ_pEX_long fl = CanZass(p,0);

	if (fl.length() != 2)
		return 0;

	ye=-coeff(fl[0].a,0);
	return 1;
}

void RF_inv(rf_elt y,rf_elt x, mplf_t f)
{
	inv(*(ZZ_pE*)y,*(ZZ_pE*)x);
}

int RF_invsqrt(rf_elt y,rf_elt x, mplf_t f)
{
	RF_inv(y,x,f);
	return RF_sqrt(y,y,f);
}

int RF_is_zero(rf_elt x, mplf_t f)
{
	return *(ZZ_pE*)x == 0;
}

void RF_mul(rf_elt z, rf_elt y,rf_elt x, mplf_t f)
{
	mul(*(ZZ_pE*)z, *(ZZ_pE*)x, *(ZZ_pE*)y);
}

#if 0
/* return x such that x^p-x=a ; 
 * such a root exists iff tr(a)==0 ; return 0 if this is not the
 * case.
 */
   
int RF_artin_schreier_root(rf_elt x,rf_elt a,mplf_t f)
{
	ZZ_pEX p;
	p=0;
	SetCoeff(p,2,1);
	SetCoeff(p,1,-1);
	SetCoeff(p,0,-*(ZZ_pE*)a);
	FindRoot(*(ZZ_pE*)x,p);
	return 1;
}


/* tells whether the trace of a is zero */
int RF_has_trace_zero(rf_elt a,mplf_t f)
{
	return trace(*(ZZ_pE*)a)==0L;
}

#endif
