/* $Id: mploc.h 186 2004-10-13 11:55:26Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef MPLOC_H_
#define MPLOC_H_

#ifdef __cplusplus
extern "C" {
#endif

#define DEBUG_MPLOC	1

#ifdef	MPLOC_ENABLE_MAGIC_NUMBERS
/* This is for an eventual C++ interface */
typedef	unsigned long int	mploc_magic_t;
#define	DECLARE_MAGIC	mploc_magic_t	magic
#define	INSERT_MAGIC(ptr,v)	ptr->magic=v
#define	MPLOC_MAGIC_Z2_ELEMENT	0x08CAF300
#define	MPLOC_MAGIC_ZP_ELEMENT	0x08CAF301
#define	MPLOC_MAGIC_Z2N_ELEMENT	0x08CAF302
#define	MPLOC_MAGIC_ZPN_ELEMENT	0x08CAF303
#define	MPLOC_MAGIC_Z2_POLY	0x08CAF304
#define	MPLOC_MAGIC_ZP_POLY	0x08CAF305
#define	MPLOC_MAGIC_Z2N_POLY	0x08CAF306
#define	MPLOC_MAGIC_ZPN_POLY	0x08CAF307
#define	MPLOC_MAGIC_FIELD	0x08CAF380
#else
#define DECLARE_MAGIC
#define INSERT_MAGIC(ptr,v)
#endif

#define mploc_precision(x)	(x)->prec

struct mplf_private;
struct mplf_struct;
struct mploc_struct;

typedef struct mplf_struct * mplf_t;
typedef struct mploc_struct * mploc_ptr;
typedef const struct mploc_struct * mploc_srcptr;

/* This struct gathers the information for the local field considered */
struct mplf_struct {
	DECLARE_MAGIC;
	mpz_t			p;
	int			degree;
	int			ramified;
	// void			(*force_precision)(mploc_ptr, mp_size_t);

	/* This has to be void * because we don't want to be bound to a
	 * particular lower level layer. The layer must provide a casting
	 * macro. Of course, the residue field might not be used at all !
	 */
	void		      * residue_field;

	struct mplf_private   * priv;
#ifdef DEBUG_MPLOC
	int			people;
#endif
};

/* This is one padic element. */
struct mploc_struct {
	DECLARE_MAGIC;
	mp_size_t	prec;	// in primes.
	struct mplf_struct * f;	// carries some useful information
	mpz_t		x;	// put last for consistency w/ ext type.
};

/* Note about the precision: It means ``known mod p^prec'' */
typedef struct mploc_struct mploc_t[1];

/************************************************************************/

struct mploc_z2x_struct;

typedef struct mploc_z2x_struct * mploc_z2x_ptr;
typedef const struct mploc_z2x_struct * mploc_z2x_srcptr;


/* This is one padic polynomial */
struct mploc_z2x_struct {
	DECLARE_MAGIC;
	mp_size_t	prec;
	struct mplf_struct * f;
	struct mplf_struct * coeff_field;
	int		deg;
	int		alloc;
	mp_size_t	cw;	/* coeff width */
	mp_limb_t	* c;
};

#ifdef	LOOSE_CW
#define	TIGHTCW(x)	iceildiv((x)->prec,GMP_LIMB_BITS)
#else
#define	TIGHTCW(x)	(x)->cw
#endif

/* Note about the precision: It means ``known mod p^prec'' */
typedef struct mploc_z2x_struct mploc_z2x_t[1];

/* Note : this data type is also used for elements of Z_{2^n}. */
typedef	mploc_z2x_ptr	mploc_z2n_ptr;
typedef mploc_z2x_srcptr	mploc_z2n_srcptr;
typedef	mploc_z2x_t	mploc_z2n_t;

/************************************************************************/

struct mploc_zpx_struct;

typedef struct mploc_zpx_struct * mploc_zpx_ptr;
typedef const struct mploc_zpx_struct * mploc_zpx_srcptr;

struct mploc_zpx_struct {
	DECLARE_MAGIC;
	mp_size_t       prec;
	struct mplf_struct * f;
	struct mplf_struct * coeff_field;
	int             deg;
	int             alloc_coeffs;  
	mpz_t		* c;
};

typedef struct mploc_zpx_struct mploc_zpx_t[1];

/* As it is, this type will do for zpx and zpn. For zpnx, an additional
 * ``cluster size'' member or equivalent could be useful
 */
typedef	mploc_zpx_ptr	mploc_zpn_ptr;
typedef mploc_zpx_srcptr	mploc_zpn_srcptr;
typedef	mploc_zpx_t	mploc_zpn_t;

/************************************************************************/

#if 0
union mploc_wrapped_struct	{
	DECLARE_MAGIC;
	mploc_t		z2;
	mploc_z2x_t	z2x;
};

typedef union mploc_wrapped_struct mploc_wrapped_t[1];
#endif

/************************************************************************/
#ifndef DIVIDE_BY_ZERO
#if __GNU_MP_VERSION >= 4
extern void __gmp_divide_by_zero (void);
#define DIVIDE_BY_ZERO()    __gmp_divide_by_zero ()
#else
extern int mploc_divide_by_zero(void);
#define DIVIDE_BY_ZERO()	mploc_divide_by_zero()
#endif
#endif

/* You may define this allocator rule to something else if you're not
 * pleased with powers of 2. Currently, we use powers of two until we
 * reach 4k, and afterwards we use multiples of 4K.
 */
#define	MALLOC_POWER_LIMIT	4096

#define	MEMSIZE_NEXT(next, need, prev)				\
	BEGIN_BLOCK						\
	next = (prev) | !(prev);				\
	if (need > MALLOC_POWER_LIMIT) {			\
		next = (need | (MALLOC_POWER_LIMIT-1))+1;	\
	} else {						\
		for(;next<(need);next<<=1);			\
	}							\
	END_BLOCK

#ifdef	TRACE_REALLOC
extern int mploc_reallocs;
extern int mploc_shuffles;
#define	LOG_REALLOC()	mploc_reallocs++
#define	LOG_SHUFFLE()	mploc_shuffles++
void REALLOC_INFO();
#else
#define	LOG_REALLOC()
#define	LOG_SHUFFLE()
#define	REALLOC_INFO()
#endif

#ifdef __cplusplus
}
#endif

#endif	/* MPLOC_H_ */
