/* $Id: z2_limbs.c 186 2004-10-13 11:55:26Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* This is all the GMP stuff I need... */
#include <gmp.h>


#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include "mploc.h"
#include "z2_limbs.h"
#include "manu.h"
#include "asm/asm.h"
#include "gmp_hacks.h"

/* All these function operate at limb level. The code in here is shared
 * between z2 and z2x.
 *
 * Of course, the whole mpn_ interface is in the same spirit. For obvious
 * reasons, we haven't duplicated all the corresponding functions.
 */
void mploc_limbs_insert_zero(mp_limb_t * p, mp_size_t n, mp_size_t k)
{
	mp_size_t i;

	i = k / GMP_LIMB_BITS;
	k = k % GMP_LIMB_BITS;

	if (k && n>i) {
		p[i] &= ((1UL<<k)-1);
		i++;
	}
	if (n>i)
		memset(p+i,0,(n-i)*sizeof(mp_limb_t));
}

/* read bit (k-1) (set to 0 if k==0). Extend up to word n including.  */
void mploc_limbs_sign_extend(mp_limb_t * p, mp_size_t n, mp_size_t k)
{
	mp_size_t i;
	mp_limb_t b;

	if (EXPECT_FALSE(k<=1)) {
		b=0UL;
	} else {
		mp_size_t u,v;
		u=(k-1) / GMP_LIMB_BITS;
		v=(k-1) % GMP_LIMB_BITS;
		b=-((p[u]&(1UL<<v))!=0);
	}

	i = k / GMP_LIMB_BITS;
	k = k % GMP_LIMB_BITS;

	if (k && n>i) {
		p[i] &= ((1UL<<k)-1);
		p[i] |= b<<k;
		i++;
	}
	if (n>i)
		memset(p+i,b,(n-i)*sizeof(mp_limb_t));
}

int mploc_limbs_are_zero(const mp_limb_t * p, mp_size_t n)
{
	int r;
	for(r=1;n--&&(r=!*p++););
	return r;
}

int mploc_limbs_are_equal(const mp_limb_t * p, const mp_limb_t * q, mp_size_t n)
{
	int r;
	for(r=1;n--&&(r=(*p++!=*q++)););
	return r;
}

/* Argh ; spent some time on this one... :-( */
void mploc_limbs_neg(mp_limb_t * d, const mp_limb_t * s, mp_size_t n)
{
	mp_size_t i;
	for(i=0;i<n&&!(d[i]=-s[i]);i++);
	for(i++;i<n;i++) d[i] = ~ s[i];
}

#define	SHORT_PRODUCT_THRESHOLD	32
#define	SHORT_SQUARE_THRESHOLD	32

/* Put the low n limbs of s1 * s2 to r ; use tmp as a buffer.
 *
 * The maximum number of limbs touched in tmp is 2n
 */
void	mploc_limbs_short_product(mp_limb_t * r,
		const mp_limb_t * s1,
		const mp_limb_t * s2,
		mp_limb_t * tmp,
		mp_size_t n)
{
	mp_size_t co;
#if 0
	mp_size_t dco;
	int i;
#endif

	ASSERT(n);
	
	if (s1==s2) {
		mploc_limbs_short_square(r,s1,tmp,n);
		return;
	}

	if (n<=8) {
		/* In case r==s, use the temporary buffer for temp.
		 * storage. This avoids the overlap problems.
		 */
		if (r==s1) {
			mploc_limbs_copy(tmp,s1,n);
			s1=tmp;
		} else if (r==s2) {
			mploc_limbs_copy(tmp,s2,n);
			s2=tmp;
		}

		switch(n)	{
			case 1: MUL1(r,s1,s2); return;
			case 2: MUL2(r,s1,s2); return;
			case 3: MUL3(r,s1,s2); return;
			case 4: MUL4(r,s1,s2); return;
			case 5: MUL5(r,s1,s2); return;
			case 6: MUL6(r,s1,s2); return;
			case 7: MUL7(r,s1,s2); return;
			case 8: MUL8(r,s1,s2); return;
		}
	}

	co = 0.694 * n;

	ASSERT(2*co >= n);

	/*
	printf("Short product %lu x %lu -> Doing Full-%lu, and 2x Short-%lu\n",
			n,n,co,n-co);
			*/

	mpn_mul_n(tmp,s1,s2,co);

	if (EXPECT_TRUE(r!=s2)) {
		mploc_limbs_short_product(r+co,s1+co,s2,tmp+n,n-co);
		/* Since we've touched r, s1[co..n[ has possibly been
		 * overwritten. Since r!=s2, it is safe to use s2[c0..n[
		 *
		 * Note that co >= n-co, to s1[0..n-co[ is safe anyway.
		 */
		mpn_add_n(tmp+co,r+co,tmp+co,n-co);
		mploc_limbs_short_product(r+co,s1,s2+co,tmp+n,n-co);
		mpn_add_n(tmp+co,r+co,tmp+co,n-co);
	} else {
		ASSERT(r!=s1);	/* otherwise s1==s2 */
		mploc_limbs_short_product(r+co,s1,s2+co,tmp+n,n-co);
		mpn_add_n(tmp+co,r+co,tmp+co,n-co);
		mploc_limbs_short_product(r+co,s1+co,s2,tmp+n,n-co);
		mpn_add_n(tmp+co,r+co,tmp+co,n-co);
	}
	mploc_limbs_copy(r,tmp,n);
}

/* No carry returned here */
void	mploc_limbs_short_square(mp_limb_t * t,
		const mp_limb_t * s,
		mp_limb_t * tmp,
		mp_size_t n)
{
	mp_size_t co;

	ASSERT(n);

	if (n<=4) {
		/* In case r==s, use the temporary buffer for temp.
		 * storage. This avoids the overlap problems.
		 */
		mploc_limbs_copy(tmp,s,n);
		switch(n)	{
			/* 20030223 : forgot to return. :-( */
			case 1: SQR1(t,tmp); return;
			case 2: SQR2(t,tmp); return;
			case 3: SQR3(t,tmp); return;
			case 4: SQR4(t,tmp); return;
		}
	}

	co = 0.694 * n;

	/* The code here works even for t == s */
	mpn_mul_n(tmp,s,s,co);

	ASSERT(2*co >= n);	/* therefore n-co <= n/2 */

	mploc_limbs_short_product(t+co,s+co,s,tmp+n,n-co);
	mpn_lshift(t+co,t+co,n-co,1);
	mploc_limbs_copy(t,tmp,co);
	mpn_add_n(t+co,t+co,tmp+co,n-co);
}

/* returns the carry bit out */
mp_limb_t	mploc_limbs_short_square_carry(mp_limb_t * t,
		const mp_limb_t * s,
		mp_limb_t * tmp,
		mp_size_t n)
{
	mp_size_t co;
	mp_limb_t c;

	ASSERT(n);

	if (n<=4) {
		/* In case t==s, use the temporary buffer for temp.
		 * storage. This avoids the overlap problems.
		 */
		mploc_limbs_copy(tmp,s,n);
		switch(n)	{
			case 1: return SQR1c(t,tmp);
			case 2: return SQR2c(t,tmp);
			case 3: return SQR3c(t,tmp);
			case 4: return SQR4c(t,tmp);
		}
	}

	co = 0.694 * n;

	/*
	printf("Short square %lu x %lu -> Doing Full-%lu, and 2x Short-%lu\n",
			n,n,co,n-co);
			*/

	mpn_mul_n(tmp,s,s,co);

	c=(2*co>n)?tmp[n]:0UL;

	ASSERT(2*co >= n);

	/* We're doing t=(l+2^c h)^2 = l^2 + 2^(c+1) * l * h, c = co * w
	 * Bit n*w is tmp[n]&1 + (tmp'[n-co]<<1)&1
	 */
	mploc_limbs_short_product(t+co,s+co,s,tmp+n,n-co);
	c^=mpn_lshift(t+co,t+co,n-co,1);
	mploc_limbs_copy(t,tmp,co);
	c^=mpn_add_n(t+co,t+co,tmp+co,n-co);
	return c&1;	/* So that it doesn't give the impression
			   to be anything else than a carry */
}

int	mploc_limbs_valuation(const mp_limb_t * pz, mp_size_t n)
{
	mp_limb_t	a;
	mp_size_t	r;
	int i;

	r=0;
	for(i=0;i<n;i++) {
		if (EXPECT((a=*pz++)!=0,1)) {
			count_trailing_zeros(r,a);
			break;
		}
	}
	if (EXPECT(i!=0,0)) {
		r+=i * GMP_LIMB_BITS;
	}
	return r;
}

