/* $Id: zpn.c 186 2004-10-13 11:55:26Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HAVE_MEMSET 1

#include <gmp.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <assert.h>

#include "gmp_hacks.h"
#include "mploc.h"
#include "zp.h"
#include "zpx.h"
#include "zpn.h"

#include "ff/rf_layer_gfpn_ntl.hpp"

#include "manu.h"

/* An element of zpn is just a polynomial.
 */

struct mplf_private {
	mploc_zpx_t	p;
	zpx_reducer_t	reducer;
	int		reducer_handles_overlaps;
};

#define	PARENT_FIELD(f)	(mplf_zpn_def_poly(f)->coeff_field)

/* mplf_zpn_create -- Create a zpn field
 *
 * INPUT: The definition polynomial
 * OUTPUT: The field structure
 */
mplf_t mplf_zpn_create(mploc_zpx_srcptr p)
{
	struct mplf_struct * f;

	f=malloc(sizeof(struct mplf_struct));
#ifdef DEBUG_MPLOC
	f->people=0;
#endif
	f->degree=p->deg;
	f->ramified=0;	/* FIXME : Obviously wrong. Perhaps this
			   could be removed, since we don't care */
	f->priv=malloc(sizeof(struct mplf_private));
	mploc_zpx_init(f->priv->p,p->coeff_field);
	mploc_zpx_set(f->priv->p,p);
	/* Default (dumb) reduction routine via Euclidean division does
	 * handle overlaps. */
	f->priv->reducer=NULL;
	f->priv->reducer_handles_overlaps=1;
	mpz_init_set(f->p,p->coeff_field->p);

	ATTACH_RESIDUE_FIELD(f);

	return f;
}

mploc_zpx_srcptr mplf_zpn_def_poly(mplf_t f)
{
	return f->priv->p;
}

/* mplf_zpn_destroy -- Frees the data asssociated with a zpn field.
 *
 * INPUT: A pointer to the field structure.
 * OUTPUT: none ; pointer is set to NULL.
 *
 * All elements attached to the field are assumed to have been cleared
 * previously.
 */
void mplf_zpn_destroy(mplf_t * pf)
{
#ifdef DEBUG_MPLOC
	ASSERT(!(*pf)->people); 
#endif
	DETACH_RESIDUE_FIELD((*pf));

	mploc_zpx_clear((*pf)->priv->p);
	mpz_clear((*pf)->p);
	free((*pf)->priv);
	free(*pf);
	*pf=NULL;
}

/* Do the X^2n / f stuff */
void mplf_zpn_set_reducer(mplf_t f, zpx_reducer_t rf, int t)
{
	f->priv->reducer=rf;
	f->priv->reducer_handles_overlaps=(rf==NULL)?1:t;
}

void mploc_zpn_reduce(mploc_zpn_ptr dst, mploc_zpn_srcptr src) 
{
	mploc_zpx_t tmp;
	mploc_zpn_ptr tmp_ptr=NULL;

	if (dst==src && !src->f->priv->reducer_handles_overlaps) {
		mploc_zpx_init(tmp,PARENT_FIELD(src->f));
		mploc_zpx_realloc(tmp,1,src->f->degree-1,src->prec);
		tmp->deg=src->f->degree-1;
		tmp_ptr=tmp;
	}

	if (src->f->priv->reducer) {
		(*(src->f->priv->reducer))(tmp_ptr?tmp_ptr:dst,src);
	} else {
		mploc_zpx_div(NULL,tmp_ptr?tmp_ptr:dst,src,src->f->priv->p);
	}

	if (tmp_ptr) {
		mploc_zpx_set(dst,tmp);
		mploc_zpx_clear(tmp);
	}
	ASSERT(mploc_zpx_is_good_looking(dst));
	ASSERT(mploc_zpn_is_reduced(dst));
}

/* These two are duplicates of the zpx analogues. But we prefer to have
 * them separated.
 */
void mploc_zpn_init(mploc_zpn_ptr x, mplf_t f)
{
	x->prec=0;
	x->deg=-1;
	x->alloc_coeffs=0;
	x->c=NULL;
	x->f=f;
	x->coeff_field=f->priv->p->coeff_field;
#ifdef DEBUG_MPLOC
	x->f->people++;
#endif
}

void mploc_zpn_clear(mploc_zpn_ptr x)
{               
#ifdef DEBUG_MPLOC
        x->f->people--;
#endif  
        free(x->c);
        memset(x,0,sizeof(mploc_zpn_ptr));
}


void mploc_zpn_mul(mploc_zpn_ptr c, mploc_zpn_srcptr a, mploc_zpn_srcptr b)
{
	/* Caveat here: if a or b have very high precision (more than the
	 * precision of the defining polynomial), some time is wasted.
	 */
	mploc_zpx_mul(c,a,b);
	mploc_zpn_reduce(c,c);
}

void mploc_zpn_sqr(mploc_zpn_ptr c, mploc_zpn_srcptr a)
{
	mploc_zpx_sqr(c,a);
	mploc_zpn_reduce(c,c);
}

void mploc_zpn_rshift(mploc_zpn_ptr z, mploc_zpn_srcptr x, mp_size_t n)
{
	if (EXPECT_FALSE(n==0)) {
		mploc_zpn_set(z,x);
	}

	if (!mploc_zpn_is_reduced(x)) {
		mploc_zpn_reduce(z,x);
	}
	/* This *has* to be so ; otherwise result is not well-defined... */
	ASSERT(mploc_zpx_valuation(x)>=n);
	mploc_zpx_rshift(z,x,n);
}

/* Note : We could do the reduction only at the very end instead...
 */
void mploc_zpn_inv_improve_recursive(mploc_zpn_ptr x, mploc_zpn_srcptr a,
		mp_size_t aim)
{
	mploc_zpx_t y;
	mploc_zpx_t t;

	if (x->prec >= aim) return;

	mploc_zpn_init(t,a->f);
	mploc_zpn_init(y,a->f);

	mploc_zpn_inv_improve_recursive(x,a,aim - aim/2);
	mploc_zpn_set_change_precision(t,a,aim);
	mploc_zpn_force_precision(x,aim);
	mploc_zpn_mul(y,t,x);
	BAD_CODE_WARNING;

	/* XXX That's a bad way of making a full product */

	mploc_zpn_sub_ui(y,y,1);
	mploc_zpn_mul(y,y,x);
	mploc_zpn_neg(y,y);
	mploc_zpn_set_change_precision(t,y,aim);
	mploc_zpn_add(x,x,t);

	x->prec=aim;

	mploc_zpn_clear(t);
	mploc_zpn_clear(y);

	ASSERT(mploc_zpx_is_good_looking(x));
}

static const char* call_reduce_insult=
	"Call _reduce() before complicated functions";

void mploc_zpn_inv(mploc_zpn_ptr z, mploc_zpn_srcptr a)
{
	rf_elt r1;
	mploc_zpx_t x;

	ASSERT(mploc_zpn_is_good_looking(a));

	if (!mploc_zpn_is_reduced(a)) {
		BASH_USER(call_reduce_insult);
		/* Otherwise we have to go through unpleaseant bloatware
		 * stuff. And to no benefit, since a is const.
		 */
	}

	if (EXPECT_FALSE(mploc_zpn_valuation(a))) DIVIDE_BY_ZERO();

	r1=new_rfe(a->f);
	rfe_set_mploc(r1,a,a->f);
	RF_inv(r1,r1,a->f);

	mploc_zpn_init(x,a->f);
	mploc_set_rfe(x,r1,a->f);
	delete_rfe(r1,a->f);
	mploc_zpn_inv_improve_recursive(x,a,a->prec);

	mploc_zpn_set(z,x);
	mploc_zpn_clear(x);
}

static void mploc_zpn_invsqrt_improve_recursive(
		mploc_zpn_ptr x,
		mploc_zpn_srcptr a,
		mp_size_t aim)
{
	mploc_zpx_t y,t;
	if (x->prec >= aim) return;
	mploc_zpn_init(t,a->f);
	mploc_zpn_init(y,a->f);
	mploc_zpn_invsqrt_improve_recursive(x,a,aim - aim/2);

	/* One must *not* change the precision of a here ... */
	mploc_zpn_set(t,a);
	ASSERT(a->prec >= aim);

	mploc_zpn_force_precision(x,aim);
	mploc_zpx_sqr(y,x);
	mploc_zpx_mul(y,y,t);
	mploc_zpn_sub_ui(y,y,1);
	mploc_zpx_mul_zp(y,y,one_half(a->coeff_field,a->prec));
	mploc_zpn_mul(y,y,x);
	mploc_zpn_set_change_precision(t,y,aim);
	mploc_zpn_sub(x,x,t);
	ASSERT(x->prec==aim);
	mploc_zpn_clear(t);
	mploc_zpn_clear(y);

	ASSERT(mploc_zpx_is_good_looking(x));
}

int mploc_zpn_invsqrt(mploc_zpn_ptr z, mploc_zpn_srcptr a)
{
	mploc_zpx_t x;
	rf_elt xbar;

	ASSERT(mploc_zpn_is_good_looking(a));

	if (!mploc_zpn_is_reduced(a)) {
		BASH_USER(call_reduce_insult);
	}

	if (EXPECT_FALSE(mploc_zpn_valuation(a))) DIVIDE_BY_ZERO();

	xbar=new_rfe(a->f);

	rfe_set_mploc(xbar,a,a->f);
	if (!RF_invsqrt(xbar,xbar,a->f)) {
		delete_rfe(xbar,a->f);
		return 0;
	}

	mploc_zpn_init(x,a->f);
	mploc_set_rfe(x,xbar,a->f);
	delete_rfe(xbar,a->f);

	mploc_zpn_invsqrt_improve_recursive(x,a,a->prec);
	mploc_zpn_set(z,x);
	mploc_zpn_clear(x);

	return 1;
}

int mploc_zpn_sqrt(mploc_zpn_ptr z, mploc_zpn_srcptr a)
{
	mp_size_t v;
	int r;

	if (a->deg==-1) {
		mploc_zpn_set_change_precision(z,a,a->prec);
		return 1;
	}

	v=mploc_zpn_valuation(a);

	if (v & 1)
		return 0;

	if (a==z || v) {
		mploc_zpn_t tmp;

		mploc_zpn_init(tmp,a->f);
		if (v) mploc_zpn_rshift(tmp,a,v);
		r=mploc_zpn_invsqrt(z,tmp);
		if (r) {
			mploc_zpn_mul(z,z,tmp);
			if (v) mploc_zpn_lshift(z,z,v/2);
		}
		mploc_zpn_clear(tmp);
	} else {
		r=mploc_zpn_invsqrt(z,a);
		if (r) {
			mploc_zpn_mul(z,z,a);
		}
	}
	return r;
}

int mploc_zpn_eq(mploc_zpn_srcptr x, mploc_zpn_srcptr y)
{
	int res;

	if (!mploc_zpn_is_reduced(y) || !mploc_zpn_is_reduced(x)) {
		mploc_zpx_t t;
		mploc_zpn_init(t,x->f);
		mploc_zpn_sub(t,x,y);
		mploc_zpn_reduce(t,t);
		res=mploc_zpx_eq_ui(t,0);
		mploc_zpn_clear(t);
	} else {
		res=mploc_zpx_eq(x,y);
	}

	return res;
}

int mploc_zpn_eq_ui(mploc_zpn_srcptr x, mp_limb_t u)
{
	int res;

	if (!mploc_zpn_is_reduced(x)) {
		mploc_zpx_t t;
		mploc_zpn_init(t,x->f);
		mploc_zpn_sub_ui(t,t,u);
		mploc_zpn_reduce(t,t);
		res=mploc_zpx_eq_ui(t,0);
		mploc_zpn_clear(t);
	} else {
		res=mploc_zpx_eq_ui(x,u);
	}

	return res;
}
